import numpy as np


class Individual:

    def __init__(self) -> None:
        self.X = None
        self.D = {}

    def get_genome(self):
        return self.X

    def set_genome(self, X):
        self.X = X


def get_genome(individuals):
    n_individuals = len(individuals)
    n_dim = individuals[0, 0].get_genome().shape[0]
    X = np.full((n_individuals, n_dim), np.inf, dtype=np.double)
    for i in range(n_individuals):
        X[i, :] = individuals[i, 0].get_genome()
    return X


def create_from_genome(clazz, X):
    if clazz is None:
        return np.full((X.shape[0], 1), np.inf)
    else:
        l = []
        for i in range(X.shape[0]):
            obj = clazz()
            obj.set_genome(X[i, :])
            l.append(obj)
        return np.array(l, dtype=np.object)[:, None]


def create_as_objects(clazz, n_objects):
    if clazz is None:
        return np.full((n_objects, 1), np.inf)
    else:
        l = []
        for i in range(n_objects):
            l.append(clazz())
        return np.array(l, dtype=np.object)[:, None]



import numpy as np
from numpy.linalg import LinAlgError


def denormalize(x, x_min, x_max):
    return x * (x_max - x_min) + x_min


def normalize(x, x_min=None, x_max=None, return_bounds=False):
    if x_min is None:
        x_min = np.min(x, axis=0)
    if x_max is None:
        x_max = np.max(x, axis=0)

    denom = x_max - x_min
    denom += 1e-30

    res = (x - x_min) / denom
    if not return_bounds:
        return res
    else:
        return res, x_min, x_max

def normalize_by_asf_interceptions__(F, non_dom, return_bounds=False):
    pass


def normalize_by_asf_interceptions_(x, return_bounds=False):

    # find the x_min point
    n_obj = x.shape[1]
    x_min = np.min(x, axis=0)
    # transform the objective that 0 means the best
    F = x - x_min

    # calculate the asf matrix
    asf = np.eye(n_obj)
    asf = asf + (asf == 0) * 1e-16

    # result matrix with the selected points
    S = np.zeros((n_obj, n_obj))

    # find for each direction the best
    for i in range(len(asf)):
        val = np.max(F / asf[i, :], axis=1) + 0.0001 * np.sum(F / asf[i, :])
        S[i, :] = F[np.argmin(val), :]

    try:
        b = np.ones(n_obj)
        A = np.linalg.solve(S, b)
        A = 1 / A
        A = A.T
    except LinAlgError:
        A = np.max(S, axis=0)

    F = F / A

    if not return_bounds:
        return F
    else:
        x_max = A + x_min
        return F, x_min, x_max


def normalize_by_asf_interceptions(X, first_front, prev_S=None, prev_asf=None, return_bounds=False):
    """

    Normalization by asf intercepts. The following is a series of checks required during the normalization procedure.

    Regarding normalization:
    At each generation we store the extreme points that are used to form hyperplane.
    If extreme points found at current generation have a smaller asf value then we use this point only and update
    the stored point, in-case stored point has smaller asf then we use stored point.

    Regarding finding intercepts:
    If the gaussian elimination fails we take nadir point as intercept value
    (nadir point is formed by taking maximum value from first front members across each dimension).
    In case gaussian elimination succeeds but some intercept value is negative then we replace that particular
    value with the corresponding nadir point value.
    Last check that we make is, if intercept value thus found is smaller than 1e-6 we replace that particular
    value with the maximum value across all population members (including last front members) in that particular dimension.

    Parameters
    ----------
    X : numpy.ndarray
        Points in the design space
    first_front :
        first_front: length of the first front, represents the first f indexes of x
    prev_S :
        Stored extreme points
    prev_asf :
        Stored asf values
    return_bounds :
        Flag to return bounds, if true, F_max an F_min of population are returned

    Returns
    -------

    pop :
        Normalized population N, asf and S matrices to store, optionally F_min, F_max

    """

    # find the x_min point
    n_obj = X.shape[1]
    # Ideal point
    x_min = np.min(X, axis=0)
    F = X - x_min

    # nadir point is calculated as the max of each objective in the FIRST front only
    nadir = np.max(F[range(first_front)], axis=0)

    # calculate the asf matrix
    asf = np.eye(n_obj)
    asf = asf + (asf == 0) * 1e-16

    # result matrix with the selected points
    S = np.zeros((n_obj, n_obj))
    new_asf = []

    # find for each direction the best
    for i in range(len(asf)):
        val = np.max(F / asf[i, :], axis=1) + 0.0001 * np.sum(F / asf[i, :])
        new_asf.append(val[np.argmin(val)])
        S[i, :] = F[np.argmin(val), :]

    # Calculate and update extreme points
    if prev_S is not None and prev_asf is not None:
        prev_asf, prev_S = recalculate(x_min, prev_asf, prev_S)
        new_asf, S = update_extreme_points(new_asf, S, prev_asf, prev_S)
    try:
        b = np.ones(n_obj)
        A = np.linalg.solve(S, b)
        # A[A==0] = 0.000001
        A = 1 / A
        A = A.T
    except LinAlgError:

        # If gaussian elimination fails we select the nadir point as the intercept
        A = nadir

    # If an intercept value is negative, replace it with the associated nadir value
    A[A < 0] = nadir[A < 0]
    # If an intercept value is smaller than 1e-6 then we replace
    # it with the max value across population members in that direction
    A[A < 1e-6] = F[:, A < 1e-6].max(axis=0)

    F = F / A

    # We undo the ideal point translation because these extreme points need to be stored in the original space
    S = S + x_min
    if not return_bounds:
        return F, new_asf, S
    else:
        x_max = A + x_min
        return F, new_asf, S, x_min, x_max


def recalculate(ideal, asf, S):
    """
    This function simply calculates the asf of a set of points S
    Variables are passed by assignment therefore the asf matrix and point matrix
    are updated in place.
    :param ideal: Ideal point of the population
    :param asf: asf matrix from previous generation
    :param S: set of points, (extreme points from previous generation)
    :return: Updates asf matrix and S matrix with new values
    """
    n_obj = len(asf)
    A = np.eye(n_obj)
    A = A + (A == 0) * 1e-16
    # retranslate extreme points to the new ideal point
    S -= ideal

    # ASF calculation and extreme point
    for i in range(len(asf)):
        val = np.max(S / A[i, :], axis=1) + 0.0001 * np.sum(S / A[i, :])
        asf[i] = val[np.argmin(val)]
        S[i, :] = S[np.argmin(val), :]
    return asf, S


def update_extreme_points(new_asf, S, prev_asf, prev_S):
    """
    Updates the extreme points for the current generation using new and stored extreme points
    :param new_asf: newly computed asf values for current generation
    :param S: newly computed extreme points for current generation
    :param prev_asf: stored asf values
    :param prev_S: stored extreme points
    :return: best asf values, and best extreme points
    """
    extreme_points = []
    min_asf = []

    for i in range(len(S)):
        if new_asf[i] < prev_asf[i]:
            extreme_points.append(S[i])
            min_asf.append(new_asf[i])
        else:
            extreme_points.append(prev_S[i])
            min_asf.append(prev_asf[i])
    return min_asf, extreme_points


def standardize(x, return_bounds=False):
    mean = np.mean(x, axis=0)
    std = np.std(x, axis=0)

    # standardize
    val = (x - mean) / std

    if not return_bounds:
        return val
    else:
        return val, mean, std


def destandardize(x, mean, std):
    return (x * std) + mean



import numpy as np

from pymoo.algorithms.nsga3 import NSGA3
from pymoo.operators.default_operators import set_if_none
from pymoo.operators.selection.tournament_selection import TournamentSelection
from pymoo.rand import random
from pymoo.util.dominator import compare


class UNSGA3(NSGA3):

    def __init__(self,
                 ref_dirs,
                 pop_size=None,
                 **kwargs):

        if pop_size is not None:
            kwargs['pop_size'] = ref_dirs.shape[0]
        set_if_none(kwargs, 'selection', TournamentSelection(func_comp=comp_by_rank_and_ref_line_dist))

        super().__init__(ref_dirs, **kwargs)


def comp_by_rank_and_ref_line_dist(pop, P, D, **kwargs):
    S = np.full(P.shape[0], np.nan)

    rank, niche = D['niche'], D['rank']

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop.CV[a, 0] > 0.0 or pop.CV[b, 0] > 0.0:
            S[i] = compare(a, pop.CV[a, 0], b, pop.CV[b, 0], method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible just set random
        else:

            # if in the same niche select by rank
            if niche[a] == niche[b]:
                S[i] = compare(a, rank[a], b, rank[b], method='smaller_is_better', return_random_if_equal=True)

            # otherwise just select random
            else:
                S[i] = random.choice([a, b])

    return S[:, None].astype(np.int)



import pickle

from pymoo.algorithms.nsga2 import NSGA2
from pymoo.model.evaluator import Evaluator
from pymop.problems.zdt import ZDT4

if __name__ == '__main__':
    n_gen = 200
    pop_size = 100

    problem = ZDT4()
    algorithm = NSGA2(pop_size=pop_size)
    evaluator = Evaluator(pop_size * n_gen)
    seed = 0

    fname = 'pynsga2_zdt4_0.dat'

    data = {'problem': problem, 'algorithm': algorithm, 'evaluator': evaluator, 'seed': seed}
    with open(fname, 'wb') as f:
        pickle.dump(data, f)



class Evaluator:
    """

    The evaluator class which is used during the algorithm execution to limit the number of evaluations.
    This can be based on convergence, maximum number of evaluations, or other criteria.

    """

    def __init__(self, n_eval=1e10):
        self.n_max_eval = n_eval
        self.n_eval = 0

    def eval(self, problem, X, **kwargs):
        """

        This function is used to return the result of one valid evaluation.

        Parameters
        ----------
        problem : class
            The problem which is used to be evaluated
        X : np.ndarray
            A two dimensional array where each row is a variable to evaluate
        kwargs : dict
            Additional arguments which might be necessary for the problem to evaluate.

        Returns
        -------
        val : any
            Returns whatever the problem used to evaluate returns. Can be only objective values or also constraints.

        """
        if len(X.shape) == 1:
            self.n_eval += 1
        else:
            self.n_eval += X.shape[0]
        return problem.evaluate(X, **kwargs)

    def n_remaining(self):
        """

        Returns
        -------
        n_remaining : int
            The number of function evaluations that are left.

        """
        return self.n_max_eval - self.n_eval

    def has_remaining(self):
        """

        Returns
        -------
        has_remaining : bool
            True if function evaluations are left and false otherwise

        """
        return self.n_eval < self.n_max_eval



from abc import abstractmethod

import numpy as np

from pymop.problem import Problem


class Sampling:
    """

    This abstract class represents any sampling strategy that can be used to create an initial population or
    an initial search point.

    """

    @abstractmethod
    def sample(self, problem, n_samples, **kwargs):
        """
        Sample new points with problem information if necessary.

        Parameters
        ----------
        problem: class
            The problem to which points should be sampled. (lower and upper bounds, discrete, binary, ...)

        n_samples: int
            Number of samples

        kwargs: class
            Any additional data that might be necessary. e.g. constants of the algorithm, ...

        Returns
        -------
        X : np.ndarray
            Samples points in a two dimensional array

        """
        pass


def sample_by_bounds(clazz, n_samples, n_var, x_min, x_max, **kwargs):
    """

    Convenience method if only the bounds are needed to create the sample points.

    Parameters
    ----------
    clazz : class
        The sampling strategy to be used.
    n_var : int
        number of variables
    n_samples : int
        number of samples
    x_min : np.array
        lower bounds
    x_max : np.array
        upper bounds
    kwargs : dict
        additional arguments

    Returns
    -------
    X : np.ndarray
        Samples points in a two dimensional array

    """

    class P(Problem):
        def __init__(self) -> None:
            self.n_var = n_var
            self.xl = np.full(n_var, x_min)
            self.xu = np.full(n_var, x_max)

    return clazz.sample(P(), n_samples, **kwargs)




import numpy as np

from pymoo.rand.random_generator import RandomGenerator
from pymoo.util.misc import swap


class CustomRandomGenerator(RandomGenerator):

    def __init__(self, seed=None):
        self.__seed = seed
        self._seed(self.__seed)

    def _seed(self, s):

        if s is None:
            import random
            self.__seed = random.random()
        elif s >= 1.0:
            import random
            random.seed(int(s))
            self.__seed = random.random()
        else:
            self.__seed = s

        self.oldrand = np.zeros(55)
        self.jrand = 0
        self.__warmup_random()

    def _rand(self, size=None):
        val = np.full(size, np.inf)
        for index, _ in np.ndenumerate(val):
            val[index] = self._rand_float()
        return val

    def perm(self, size):
        a = np.arange(size)
        for i in range(size):
            rnd = self.randint(i, size)
            swap(a, rnd, i)
        return a

    def _rand_float(self):
        self.jrand += 1
        if self.jrand >= 55:
            self.jrand = 1
            self.__advance_random()
        return self.oldrand[self.jrand]

    def __warmup_random(self):
        self.oldrand[54] = self.__seed
        new_random = 0.000000001
        prev_random = self.__seed
        for j1 in range(1, 55):
            ii = (21 * j1) % 54
            self.oldrand[ii] = new_random
            new_random = prev_random - new_random
            if new_random < 0.0:
                new_random += 1.0
            prev_random = self.oldrand[ii]

        self.__advance_random()
        self.__advance_random()
        self.__advance_random()
        self.jrand = 0

    def __advance_random(self):
        for j1 in range(24):
            new_random = self.oldrand[j1] - self.oldrand[j1 + 31]
            if new_random < 0.0:
                new_random = new_random + 1.0
            self.oldrand[j1] = new_random

        for j1 in range(24, 55):
            new_random = self.oldrand[j1] - self.oldrand[j1 - 24]
            if new_random < 0.0:
                new_random = new_random + 1.0
            self.oldrand[j1] = new_random



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation


def plot(F, problem=None):

    if F.ndim == 1:
        print("Cannot plot a one dimensional array.")
        return

    n_dim = F.shape[1]

    if n_dim == 2:
        plot_2d(F, problem)
    elif n_dim == 3:
        plot_3d(F)
    else:
        print("Cannot plot a %s dimensional array." % n_dim)
        return


def plot_3d(F):
    fig = plt.figure()
    from mpl_toolkits.mplot3d import Axes3D
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(F[:, 0], F[:, 1], F[:, 2])

def plot_2d(F, problem=None):
    plt.scatter(F[:, 0], F[:, 1], label="F")

    if problem is not None:
        pf = problem.pareto_front()
        if pf is not None:
            plt.scatter(pf[:, 0], pf[:, 1], label='Pareto Front', s=20, facecolors='none', edgecolors='r')
            plt.legend()


def animate(path_to_file, history, problem=None, func_iter=None):

    H = np.concatenate([e['pop'].F[None, :, :] for e in history], axis=0)

    if H.ndim != 3 or H.shape[2] != 2:
        print("Can only animate a two dimensional set of arrays.")
        return

    fig = plt.figure()
    ax = plt.gca()

    # plot the pareto front if it is known for the problem
    if problem is not None:
        pf = problem.pareto_front()
        plt.scatter(pf[:, 0], pf[:, 1], label='Pareto Front', s=60, facecolors='none', edgecolors='r')

    # plot the initial population
    _F = H[0, :, :]
    scat = plt.scatter(_F[:, 0], _F[:, 1])
    plt.title("0")

    if func_iter is not None:
        func_iter(ax, history[0])


    # the update method
    def update(n):
        _F = H[n, :, :]
        scat.set_offsets(_F)

        # get the bounds for plotting and add padding
        min = np.min(_F, axis=0) - 0.1
        max = np.max(_F, axis=0) + 0.1

        # set the scatter object with padding
        ax.set_xlim(min[0], max[0])
        ax.set_ylim(min[1], max[1])

        if func_iter is not None:
            func_iter(ax, history[n])

        plt.title(n)

    # create the animation
    ani = animation.FuncAnimation(fig, update, frames=H.shape[0])

    # write the file
    Writer = animation.writers['ffmpeg']
    writer = Writer(fps=6, bitrate=1800)
    ani.save(path_to_file, writer=writer)






import numpy as np
from scipy import special

from pymoo.util.misc import unique_rows
from pymoo.util.plotting import plot_3d


class ReferenceDirectionFactory:

    def __init__(self, n_dim, scaling=None) -> None:
        super().__init__()
        self.n_dim = n_dim
        self.scaling = scaling

    def do(self):

        if self.n_dim == 1:
            return np.array([[1.0]])
        else:
            ref_dirs = self._do()
            if self.scaling is not None:
                ref_dirs = ref_dirs * self.scaling + ((1 - self.scaling) / self.n_dim)
            return ref_dirs

    def _do(self):
        return None


class UniformReferenceDirectionFactory(ReferenceDirectionFactory):

    def __init__(self, n_dim, scaling=None, n_points=None, n_partitions=None) -> None:

        super().__init__(n_dim, scaling=scaling)

        if n_points is not None:
            self.n_partitions = UniformReferenceDirectionFactory.get_partition_closest_to_points(n_points, n_dim)
        else:
            if n_partitions is None:
                raise Exception("Either provide number of partitions or number of points.")
            else:
                self.n_partitions = n_partitions

    def _do(self):
        return self.uniform_reference_directions(self.n_partitions, self.n_dim)

    @staticmethod
    def get_partition_closest_to_points(n_points, n_dim):

        # in this case the do method will always return one values anyway
        if n_dim == 1:
            return 0

        n_partitions = 1
        _n_points = UniformReferenceDirectionFactory.get_n_points(n_partitions, n_dim)
        while _n_points <= n_points:
            n_partitions += 1
            _n_points = UniformReferenceDirectionFactory.get_n_points(n_partitions, n_dim)
        return n_partitions - 1

    @staticmethod
    def get_n_points(n_partitions, n_dim):
        return int(special.binom(n_dim + n_partitions - 1, n_partitions))

    def uniform_reference_directions(self, n_partitions, n_dim):
        ref_dirs = []
        ref_dir = np.full(n_dim, np.inf)
        self.__uniform_reference_directions(ref_dirs, ref_dir, n_partitions, n_partitions, 0)
        return np.concatenate(ref_dirs, axis=0)

    def __uniform_reference_directions(self, ref_dirs, ref_dir, n_partitions, beta, depth):
        if depth == len(ref_dir) - 1:
            ref_dir[depth] = beta / (1.0 * n_partitions)
            ref_dirs.append(ref_dir[None, :])
        else:
            for i in range(beta + 1):
                ref_dir[depth] = 1.0 * i / (1.0 * n_partitions)
                self.__uniform_reference_directions(ref_dirs, np.copy(ref_dir), n_partitions, beta - i,
                                                    depth + 1)


class MultiLayerReferenceDirectionFactory:

    def __init__(self, layers=[]) -> None:
        self.layers = layers

    def add_layer(self, factory):
        self.layers.append(factory)

    def do(self):
        ref_dirs = []
        for factory in self.layers:
            ref_dirs.append(factory.do())

        ref_dirs = np.concatenate(ref_dirs, axis=0)

        return unique_rows(ref_dirs)


if __name__ == '__main__':
    ref_dirs = UniformReferenceDirectionFactory(2, n_points=100).do()
    print(np.sum(ref_dirs, axis=1))

    multi_layer = MultiLayerReferenceDirectionFactory()
    multi_layer.add_layer(UniformReferenceDirectionFactory(3, n_partitions=12, scaling=0.5))
    multi_layer.add_layer(UniformReferenceDirectionFactory(3, n_partitions=12, scaling=1.0))

    # multi_layer.add_layer(0.5, UniformReferenceDirectionFactory(3, n_partitions=10))
    ref_dirs = multi_layer.do()

    print(UniformReferenceDirectionFactory.get_partition_closest_to_points(100, 3))
    print(ref_dirs.shape)

    plot_3d(ref_dirs)



import os
import pickle
import sys
import time

import numpy as np

if __name__ == '__main__':
    """
    This class performs one run given a pickled object saving all the necessary data.
    """

    # insert this project in path
    sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

    import sys

    sys.path.append("/home/blankjul/workspace/pymop/")
    sys.path.append("/Users/julesy/workspace/pymop/")

    import pymop

    if len(sys.argv) < 2:
        raise Exception("Usage: python run.py <dat> [<out>]")

    # load the data for the experiment
    fname = sys.argv[1]
    with open(fname, 'rb') as f:
        data = pickle.load(f)

    problem, algorithm, termination, seed = data['problem'], data['algorithm'], data['termination'], data['seed']

    if len(sys.argv) == 2:
        tmp = sys.argv[1]
        out = tmp[:tmp.find('.')] + '.out'
    else:
        out = sys.argv[2]

    start_time = time.time()
    try:

        res = algorithm.solve(problem,
                              termination=termination,
                              seed=seed,
                              save_history=(len(sys.argv) == 4),
                              return_only_feasible=False,
                              return_only_non_dominated=False,
                              disp=False)

        F = res['F']

        elapsed = (time.time() - start_time)
        print(fname, "in --- %s seconds ---" % elapsed)

        # create directory of necessary
        os.makedirs(os.path.dirname(out), exist_ok=True)

        np.savetxt(out, F)

        if len(sys.argv) == 4:
            out_dat = sys.argv[3]

            from pymoo.util.non_dominated_sorting import NonDominatedSorting
            from pymoo.indicators.igd import IGD

            hist = res['history']
            pf = problem.pareto_front()

            igd = np.zeros(len(hist))

            nadir_point = np.zeros((len(hist), problem.n_obj))
            ideal_point = np.zeros((len(hist), problem.n_obj))

            for i, e in enumerate(hist):

                nadir_point[i, :] = e.survival.nadir_point
                ideal_point[i, :] = e.survival.ideal_point

                F = e.D['pop'].F
                I = NonDominatedSorting().do(F, only_non_dominated_front=True)
                igd[i] = IGD(pf).calc(F[I, :])

            D = {'igd': igd, 'nadir_point': nadir_point, 'ideal_point': ideal_point}

            pickle.dump(D, open(out_dat, 'wb'))

    except Exception as e:
        print(e)
        print("Error: %s" % fname)

    # pf = problem.pareto_front()
    # print(IGD(pf).calc(F))
    # print(IGD(pf).calc(_F))



import abc

import numpy as np


class Indicator:

    def __init__(self):
        self.default_if_empty = 0.0

    def calc(self, F):

        # if it is a 1d array
        if len(F.shape) == 1:
            F = np.array([F])

        # if there are not data
        if F.shape[1] == 0:
            return self.default_if_empty

        return self._calc(F)

    @abc.abstractmethod
    def _calc(self, F):
        return





import numpy as np

from pymoo.model.mutation import Mutation
from pymoo.rand import random


class PolynomialMutation(Mutation):
    def __init__(self, eta_mut, prob_mut=None):
        self.eta_mut = float(eta_mut)
        if prob_mut is not None:
            self.prob_mut = float(prob_mut)
        else:
            self.prob_mut = None

    def _do(self, p, X, Y, **kwargs):

        if self.prob_mut is None:
            self.prob_mut = 1.0 / p.n_var

        do_mutation = random.random(X.shape) < self.prob_mut

        Y[:, :] = X

        xl = np.repeat(p.xl[None, :], X.shape[0], axis=0)[do_mutation]
        xu = np.repeat(p.xu[None, :], X.shape[0], axis=0)[do_mutation]

        X = X[do_mutation]

        delta1 = (X - xl) / (xu - xl)
        delta2 = (xu - X) / (xu - xl)

        mut_pow = 1.0 / (self.eta_mut + 1.0)

        rand = random.random(X.shape)
        mask = rand <= 0.5
        mask_not = np.logical_not(mask)

        deltaq = np.zeros(X.shape)

        xy = 1.0 - delta1
        val = 2.0 * rand + (1.0 - 2.0 * rand) * (np.power(xy, (self.eta_mut + 1.0)))
        d = np.power(val, mut_pow) - 1.0
        deltaq[mask] = d[mask]

        xy = 1.0 - delta2
        val = 2.0 * (1.0 - rand) + 2.0 * (rand - 0.5) * (np.power(xy, (self.eta_mut + 1.0)))
        d = 1.0 - (np.power(val, mut_pow))
        deltaq[mask_not] = d[mask_not]

        # mutated values
        _Y = X + deltaq * (xu - xl)

        # back in bounds if necessary (floating point issues)
        _Y[_Y < xl] = xl[_Y < xl]
        _Y[_Y > xu] = xu[_Y > xu]

        # set the values for output
        Y[do_mutation] = _Y




import numpy as np

from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.operators.crossover.real_simulated_binary_crossover import SimulatedBinaryCrossover
from pymoo.operators.default_operators import set_if_none
from pymoo.operators.mutation.real_polynomial_mutation import PolynomialMutation
from pymoo.operators.sampling.real_random_sampling import RealRandomSampling
from pymoo.operators.selection.tournament_selection import TournamentSelection
from pymoo.operators.survival.fitness_survival import FitnessSurvival
from pymoo.util.display import disp_single_objective
from pymoo.util.dominator import compare


class SingleObjectiveGeneticAlgorithm(GeneticAlgorithm):

    def __init__(self, **kwargs):

        set_if_none(kwargs, 'sampling', RealRandomSampling())
        set_if_none(kwargs, 'selection', TournamentSelection(func_comp=comp_by_cv_and_fitness))
        set_if_none(kwargs, 'crossover', SimulatedBinaryCrossover(prob_cross=0.9, eta_cross=3))
        set_if_none(kwargs, 'mutation', PolynomialMutation(prob_mut=None, eta_mut=5))
        set_if_none(kwargs, 'survival', FitnessSurvival())
        set_if_none(kwargs, 'eliminate_duplicates', True)

        super().__init__(**kwargs)
        self.func_display_attrs = disp_single_objective


def comp_by_cv_and_fitness(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop.CV[a, 0] > 0.0 or pop.CV[b, 0] > 0.0:
            S[i] = compare(a, pop.CV[a, 0], b, pop.CV[b, 0], method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible just set random
        else:
            S[i] = compare(a, pop.F[a, 0], b, pop.F[b, 0], method='smaller_is_better', return_random_if_equal=True)

    return S[:, None].astype(np.int)



import math

import numpy as np

from pymoo.model.individual import Individual
from pymoo.model.individual import get_genome, create_from_genome, create_as_objects
from pymoo.model.algorithm import Algorithm
from pymoo.model.evaluator import Evaluator
from pymoo.model.population import Population
from pymoo.util.misc import cdist


class GeneticAlgorithm(Algorithm):
    """
    This class represents a basic genetic algorithm that can be extended and modified by
    providing different modules or operators.

    Attributes
    ----------

    pop_size: int
        The population size for the genetic algorithm. Depending on the problem complexity and modality the
        it makes sense to experiment with the population size.
        Also, to create a steady state algorithm the offspring_size can be changed.

    sampling : class or numpy.array
        The sampling methodology that is used to create the initial population in the first generation. Also,
        the initial population can be provided directly in case it is known deterministically beforehand.

    selection : pymoo.model.selection.Selection
        The mating selection methodology that is used to determine the parents for the mating process.

    crossover : pymoo.model.selection.Crossover
        The crossover methodology that recombines at least two parents to at least one offspring. Depending on
        the arity the number of crossover execution might vary.

    mutation : pymoo.model.selection.Mutation
        The mutation methodology that is used to perturbate an individual. After performing the crossover
        a mutation is executed.

    survival : pymoo.model.selection.Survival
        Each generation usually a survival selection is performed to follow the survival of the fittest principle.
        However, other strategies such as niching, diversity preservation and so on can be implemented here.

    n_offsprings : int
        Number of offsprings to be generated each generation. Can be 1 to define a steady-state algorithm.
        Default it is equal to the population size.

    """

    def __init__(self,
                 pop_size,
                 sampling,
                 selection,
                 crossover,
                 mutation,
                 survival,
                 n_offsprings=None,
                 eliminate_duplicates=False,
                 func_repair=None,
                 clazz=Individual().__class__,
                 **kwargs
                 ):

        super().__init__(**kwargs)

        self.pop_size = pop_size
        self.sampling = sampling
        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation
        self.survival = survival
        self.n_offsprings = n_offsprings
        self.eliminate_duplicates = eliminate_duplicates
        self.func_repair = func_repair
        self.clazz = clazz
        self.D = {}

        # default set the number of offsprings to the population size
        if self.n_offsprings is None:
            self.n_offsprings = pop_size

    def _next(self, pop):

        # do the iteration for the next generation - population objective is modified inplace
        # do the mating and evaluate the offsprings
        off = self._mating(pop)

        # if it is a custom class also create the custom object and store it in the population
        off.individuals = create_from_genome(self.clazz, off.X)

        off.F, off.CV, off.G = self.D['evaluator'].eval(self.D['problem'], off.X, D=self.D, individuals=off.individuals,
                                                        return_constraint_violation=True, return_constraints=True)
        self.D = {**self.D, 'off': off}

        # do the survival selection with the merged population
        pop.merge(off)
        self.survival.do(pop, self.pop_size, D=self.D)

        return off

    def _solve(self, problem, termination):

        # the evaluator object which is counting the evaluations
        evaluator = Evaluator()

        # generation counter
        n_gen = 1

        # always create a new function evaluator which is counting the evaluations
        self.D = {**self.D, 'problem': problem, 'evaluator': evaluator, 'n_gen': n_gen}

        # initialize the first population and evaluate it
        pop = self._initialize()
        self.D = {**self.D, 'pop': pop}
        self._each_iteration(self.D, first=True)

        # while termination criterium not fulfilled
        while termination.do_continue(self.D):
            self.D['n_gen'] += 1

            # do the next iteration
            self._next(pop)

            # execute the callback function in the end of each generation
            self._each_iteration(self.D)

        self._finalize()

        return pop

    def _initialize(self):

        problem, evaluator = self.D['problem'], self.D['evaluator']

        pop = Population()

        # add to the data dictionary to be used in all modules
        self.D = {**self.D, 'pop': pop}

        if isinstance(self.sampling, np.ndarray):
            pop.X = self.sampling
        else:
            pop.X = self.sampling.sample(problem, self.pop_size, D=self.D)

        # if we got a custom object during the sampling - figure out the clazz type directly
        if pop.X.dtype == np.object:

            # set the class of the custom object to be used for factory later
            self.clazz = pop.X[0, 0].__class__
            obj = pop.X[0, 0].get_genome()

            # now either they return also float - we extract the matrix
            if obj is not None and obj.dtype == np.float:
                pop.individuals = pop.X
                pop.X = get_genome(pop.individuals)

        # the object has a type which can be represented in a matrix - only if clazz is not None create an extra object
        else:

            # if no individual object is desired create a dummy array
            if self.clazz is None:
                pop.individuals = np.full((self.pop_size, 1), np.inf)

            # otherwise we create objects from this clazz
            else:
                pop.individuals = create_as_objects(self.clazz, self.pop_size)

        # add to the data dictionary to be used in all modules
        self.D = {**self.D, 'pop': pop}

        pop.F, pop.CV, pop.G = evaluator.eval(problem, pop.X, D=self.D, individuals=pop.individuals,
                                              return_constraint_violation=True, return_constraints=True)

        # that call is a dummy survival to set attributes that are necessary for the mating selection
        self.survival.do(pop, self.pop_size, D=self.D)

        return pop

    def _mating(self, pop):

        off = Population(X=np.full((0, self.D['problem'].n_var), np.inf), individuals=create_as_objects(self.clazz, 0))
        n_gen_off = 0
        n_parents = self.crossover.n_parents

        # mating counter - counts how often the mating needs to be done to fill up n_offsprings
        n_matings = 0

        # do the mating until all offspring are created usually it should be done is one iteration
        # through duplicate eliminate more might be necessary
        while n_gen_off < self.n_offsprings:

            # select from the current population individuals for mating
            n_select = math.ceil((self.n_offsprings - n_gen_off) / self.crossover.n_children)
            parents = self.selection.do(pop, n_select, n_parents, D=self.D, indviduals=pop.individuals)

            # object that represent the individuals
            _individuals = create_as_objects(self.clazz, parents.shape[0] * self.crossover.n_children)

            _X = self.crossover.do(self.D['problem'], pop.X[parents.T], D=self.D, individuals=_individuals)

            # do the mutation
            _X = self.mutation.do(self.D['problem'], _X, D=self.D, individuals=_individuals)

            # repair the individuals if necessary
            if self.func_repair is not None:
                self.func_repair(self.D['problem'], _X, D=self.D, individuals=_individuals)

            # eliminate duplicates if too close to the current population
            if self.eliminate_duplicates:
                not_equal = np.where(np.all(cdist(_X, pop.X) > 1e-12, axis=1))[0]
                _X, _individuals = _X[not_equal, :], _individuals[not_equal, :]

            # if more offsprings than necessary - truncate them
            if _X.shape[0] > self.n_offsprings - n_gen_off:
                _X = _X[:self.n_offsprings - n_gen_off, :]
                _individuals = _individuals[:self.n_offsprings - n_gen_off, :]

            # add to the offsprings
            off.merge(Population(X=_X, individuals=_individuals))
            n_gen_off += _X.shape[0]

            # increase the mating number
            n_matings += 1

            # if no new offsprings can be generated within 100 trails -> return the current result
            if n_matings > 100:
                self.evaluator.n_eval = self.evaluator.n_max_eval
                break

        return off

    def _finalize(self):
        pass



import numpy as np
from numpy.linalg import LinAlgError

from pymoo.cython.my_math import cython_calc_perpendicular_distance
from pymoo.model.survival import Survival, split_by_feasibility
from pymoo.operators.survival.reference_line_survival import niching, calc_niche_count
from pymoo.util.mathematics import Mathematics
from pymoo.util.non_dominated_sorting import NonDominatedSorting


class ReferenceLineSurvival(Survival):
    def __init__(self, ref_dirs):
        super().__init__()
        self.ref_dirs = ref_dirs
        self.extreme_points = None
        self.intercepts = None
        self.nadir_point = None
        self.ideal_point = np.full(ref_dirs.shape[1], np.inf)

    def _do(self, pop, n_survive, D=None, **kwargs):

        # convert to integer for later usage
        n_survive = int(n_survive)

        # first split by feasibility for normalization
        feasible, infeasible = split_by_feasibility(pop)

        # number of survivors from the feasible population 
        # in case of having not enough feasible solution all feasible will survive
        if len(feasible) < n_survive:
            n_survive_feasible = len(feasible)
        else:
            n_survive_feasible = n_survive

        # attributes to be set after the survival
        survivors, rank, niche_of_individuals, dist_to_niche = [], [], [], []

        # if there are feasible solutions to survive
        if len(feasible) > 0:

            # consider only feasible solutions form now on
            F = pop.F[feasible, :]

            # find or usually update the new ideal point - from feasible solutions
            self.ideal_point = np.min(np.concatenate([self.ideal_point[None, :], F], axis=0), axis=0)

            # calculate the fronts of the population
            fronts, _rank = NonDominatedSorting(epsilon=Mathematics.EPS).do(F, return_rank=True,
                                                                            n_stop_if_ranked=n_survive_feasible)
            non_dominated, last_front = fronts[0], fronts[-1]

            # calculate the worst point of feasible individuals
            worst_point = np.max(F, axis=0)
            # calculate the nadir point from non dominated individuals
            nadir_point = np.max(F[non_dominated, :], axis=0)

            # find the extreme points for normalization
            self.extreme_points = get_extreme_points_c(F, self.ideal_point, extreme_points=self.extreme_points)

            # find the intercepts for normalization and do backup if gaussian elimination fails
            self.intercepts = get_intercepts_c(self.extreme_points, self.ideal_point, nadir_point, worst_point)
            self.nadir_point = self.ideal_point + self.intercepts

            # index of the first n fronts form now on - including splitting front
            I = np.concatenate(fronts)
            F = F[I, :]

            # associate individuals to niches
            niche_of_individuals, dist_to_niche = c_associate_to_niches(F, self.ref_dirs, self.ideal_point,
                                                                      self.nadir_point)

            # if a splitting of the last front is not necessary
            if F.shape[0] == n_survive_feasible:
                _survivors = np.arange(F.shape[0])

            # otherwise we have to select using niching
            else:

                _last_front = np.arange(len(I) - len(last_front), len(I))
                _until_last_front = np.arange(0, len(I) - len(last_front))

                _survivors = []
                n_remaining = n_survive_feasible
                niche_count = np.zeros(len(self.ref_dirs), dtype=np.int)

                if len(fronts) > 1:
                    _survivors.extend(_until_last_front)
                    niche_count = calc_niche_count(len(self.ref_dirs), niche_of_individuals[_until_last_front])
                    n_remaining -= len(_until_last_front)

                S = niching(F[_last_front, :], n_remaining, niche_count, niche_of_individuals[_last_front],
                            dist_to_niche[_last_front])

                # S = niching_vectorized(F[_last_front, :], n_remaining, niche_count, niche_of_individuals[_last_front],
                #                       dist_to_niche[_last_front])

                _survivors.extend(_last_front[S].tolist())

            # reindex the survivors to the absolute index
            survivors = feasible[I[_survivors]]

            # save the attributes for surviving individuals
            rank = _rank[I[_survivors]]
            niche_of_individuals = niche_of_individuals[_survivors]
            dist_to_niche = dist_to_niche[_survivors]

        # if we need to fill up with infeasible solutions - we do so. Also, the data structured need to be reindexed
        n_infeasible = n_survive - len(survivors)
        if n_infeasible > 0:
            survivors = np.concatenate([survivors, infeasible[:n_infeasible]])
            rank = np.concatenate([rank, Mathematics.INF * np.ones(n_infeasible)])
            niche_of_individuals = np.concatenate([niche_of_individuals, -1 * np.ones(n_infeasible)])
            dist_to_niche = np.concatenate([dist_to_niche, Mathematics.INF * np.ones(n_infeasible)])

        # set attributes globally for other modules
        if D is not None:
            D['rank'] = rank
            D['niche'] = niche_of_individuals
            D['dist_to_niche'] = dist_to_niche

        # now truncate the population
        pop.filter(survivors)


def get_extreme_points_c(F, ideal_point, extreme_points=None):
    # calculate the asf which is used for the extreme point decomposition
    asf = np.eye(F.shape[1])
    asf[asf == 0] = 1e-6

    # add the old extreme points to never loose them for normalization
    _F = F
    if extreme_points is not None:
        _F = np.concatenate([_F, extreme_points], axis=0)

    # use __F because we substitute small values to be 0
    __F = _F - ideal_point
    __F[__F < 1e-3] = 0

    # update the extreme points for the normalization having the highest asf value each
    F_asf = np.max(__F / asf[:, None, :], axis=2)
    I = np.argmin(F_asf, axis=1)
    extreme_points = _F[I, :]

    return extreme_points


def get_intercepts_c(extreme_points, ideal_point, nadir_point, worst_point):
    # normalization of the points in the new space
    nadir_point -= ideal_point
    worst_point -= ideal_point

    use_nadir = False

    try:
        # find the intercepts using gaussian elimination
        plane = np.linalg.solve(extreme_points - ideal_point, np.ones(extreme_points.shape[1]))

        # if the plane
        if np.any(plane <= 1e-6):
            use_nadir = True

    except LinAlgError:
        use_nadir = True

    if use_nadir:
        intercepts = nadir_point
    else:
        intercepts = 1 / plane

    # if also the worst point is very small we set it to a small value, to avoid division by zero
    b = intercepts < 1e-6
    intercepts[b] = worst_point[b]

    b = intercepts < 1e-6
    intercepts[b] = 1e-12

    return intercepts


def c_associate_to_niches(F, niches, ideal_point, nadir_point):
    # normalize by ideal point and intercepts
    N = (F - ideal_point) / (nadir_point - ideal_point)

    # dist_matrix = calc_perpendicular_dist_matrix(N, niches)
    dist_matrix = cython_calc_perpendicular_distance(N, niches)

    niche_of_individuals = np.argmin(dist_matrix, axis=1)
    dist_to_niche = dist_matrix[np.arange(F.shape[0]), niche_of_individuals]

    return niche_of_individuals, dist_to_niche





import copy
from abc import abstractmethod

import numpy as np

from pymoo.model.termination import MaximumFunctionCallTermination
from pymoo.rand import random
from pymoo.util.non_dominated_sorting import NonDominatedSorting


class Algorithm:
    """

    This class represents the abstract class for any algorithm to be implemented. Most importantly it
    provides the solve method that is used to optimize a given problem.

    The solve method provides a wrapper function which does validate the input.

    """

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.disp = None
        self.func_display_attrs = None
        self.callback = None
        self.history = []

    def solve(self,
              problem,
              termination,
              seed=1,
              disp=False,
              callback=None,
              return_only_feasible=True,
              return_only_non_dominated=True,
              save_history=False,
              **kwargs
              ):
        """

        Solve a given problem by a given evaluator. The evaluator determines the termination condition and
        can either have a maximum budget, hypervolume or whatever. The problem can be any problem the algorithm
        is able to solve.

        Parameters
        ----------

        problem: class
            Problem to be solved by the algorithm

        evaluator: class
            object that evaluates and saves the number of evaluations and determines the stopping condition

        seed: int
            Random seed for this run. Before the algorithm starts this seed is set.

        disp : bool
            If it is true than information during the algorithm execution are displayed

        callback : func
            A callback function can be passed that is executed every generation. The parameters for the function
            are the algorithm itself, the number of evaluations so far and the current population.

                def callback(algorithm):
                    pass

        return_only_feasible : bool
            If true, only feasible solutions are returned.

        return_only_non_dominated : bool
            If true, only the non dominated solutions are returned. Otherwise, it might be - dependend on the
            algorithm - the final population

        save_history : bool
            If true, a current snapshot of each generation is saved.

        Returns
        -------
        res : dict
            A dictionary that saves all the results of the algorithm. Also, the history if save_history is true.

        """

        # set the random seed for generator
        random.seed(seed)

        self.disp = disp
        self.callback = callback
        self.save_history = save_history

        if isinstance(termination, int):
            termination = MaximumFunctionCallTermination(termination)

        # call the algorithm to solve the problem
        pop = self._solve(problem, termination, **kwargs)

        if return_only_feasible:
            feasible = np.where(pop.CV[:, 0] <= 0.0)[0]
            pop.filter(feasible)

        if pop.size() > 0 and return_only_non_dominated:
            I = NonDominatedSorting().do(pop.F, only_non_dominated_front=True)
            pop.filter(I)

        res = {'problem': problem, 'history': self.history}
        res = {**res, **pop.D}

        return res

    # method that is called each iteration to call so#me methods regularly
    def _each_iteration(self, D, first=False, **kwargs):

        # display the output if defined by the algorithm
        if self.disp and self.func_display_attrs is not None:
            disp = self.func_display_attrs(D['problem'], D['evaluator'], D)
            if disp is not None:
                self._display(disp, header=first)

        # if a callback function is provided it is called after each iteration
        if self.callback is not None:
            self.callback(self)

        if self.save_history:
            hist = self.history
            self.history = None

            obj = copy.deepcopy(self)
            self.history = hist

            self.history.append(obj)

    # attributes are a list of tuples of length 3: (name, val, width)
    def _display(self, disp, header=False):
        regex = " | ".join(["{}"] * len(disp))
        if header:
            print("=" * 40)
            print(regex.format(*[name.ljust(width) for name, _, width in disp]))
            print("=" * 40)
        print(regex.format(*[str(val).ljust(width) for _, val, width in disp]))

    @abstractmethod
    def _solve(self, problem, termination):
        pass




import numpy as np

from pymoo.algorithms.nsga3 import NSGA3
from pymoo.operators.survival.aspiration_point_survival import AspirationPointSurvival
from pymoo.util.reference_direction import UniformReferenceDirectionFactory


class RNSGA3(NSGA3):

    def __init__(self,
                 ref_points,
                 pop_per_ref_point,
                 mu=0.1,
                 **kwargs):

        n_obj = ref_points.shape[1]
        n_ref_points = ref_points.shape[0]

        # add the aspiration point lines
        aspiration_ref_dirs = []
        for i in range(n_ref_points):
            ref_dirs = UniformReferenceDirectionFactory(n_dim=n_obj, n_points=pop_per_ref_point).do()
            aspiration_ref_dirs.extend(ref_dirs)
        aspiration_ref_dirs = np.array(aspiration_ref_dirs)

        kwargs['ref_dirs'] = aspiration_ref_dirs
        super().__init__(**kwargs)

        # create the survival strategy
        self.survival = AspirationPointSurvival(ref_points, aspiration_ref_dirs, mu=mu)

    def _solve(self, problem, termination):

        if self.survival.ref_points.shape[1] != problem.n_obj:
            raise Exception("Dimensionality of reference points must be equal to the number of objectives: %s != %s" %
                            (self.survival.ref_points.shape[1], problem.n_obj))

        return super()._solve(problem, termination)


import random

from pymoo.rand.random_generator import RandomGenerator


class SecureRandomGenerator(RandomGenerator):

    def _seed(self, n):
        random.SystemRandom().seed(n)

    def _rand_float(self):
        return random.SystemRandom().random()







import numpy as np

from pymoo.dm.decision_making import DecisionMaking


class PseudoWeights(DecisionMaking):

    def __init__(self, weights) -> None:
        super().__init__()
        self.weights = weights
        self.pseudo_weights = None

    def do(self, F, **kwargs):
        # get minimum and maximum for each objective
        F_min = np.min(F, axis=0)
        F_max = np.max(F, axis=0)

        # calculate the norm for each objective
        norm = F_max - F_min

        # normalized distance to the worst solution
        self.pseudo_weights = ((F_max - F) / norm)

        # normalize weights to sum up to one
        self.pseudo_weights = self.pseudo_weights / np.sum(self.pseudo_weights, axis=1)[:, None]

        # normalize the provided weights as well
        weights_norm = self.weights / np.sum(self.weights)

        # search for the closest individual having this pseudo weights
        I = np.argmin(np.sum(np.abs(self.pseudo_weights - weights_norm), axis=1))

        return I


if __name__ == '__main__':
    pw = PseudoWeights(np.array([0.1, 0.4]))

    pw.do(np.random.rand(100, 2))



import numpy as np

from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.operators.crossover.real_simulated_binary_crossover import SimulatedBinaryCrossover
from pymoo.operators.default_operators import set_if_none
from pymoo.operators.mutation.real_polynomial_mutation import PolynomialMutation
from pymoo.operators.sampling.real_random_sampling import RealRandomSampling
from pymoo.operators.selection.tournament_selection import TournamentSelection
from pymoo.operators.survival.c_reference_line_survival import ReferenceLineSurvival
from pymoo.rand import random
from pymoo.util.display import disp_multi_objective
from pymoo.util.dominator import compare


class NSGA3(GeneticAlgorithm):

    def __init__(self, ref_dirs, **kwargs):

        self.ref_dirs = ref_dirs
        set_if_none(kwargs, 'pop_size', len(ref_dirs))
        set_if_none(kwargs, 'sampling', RealRandomSampling())
        set_if_none(kwargs, 'selection', TournamentSelection(func_comp=comp_by_cv_then_random))
        set_if_none(kwargs, 'crossover', SimulatedBinaryCrossover(prob_cross=1.0, eta_cross=20))
        set_if_none(kwargs, 'mutation', PolynomialMutation(prob_mut=None, eta_mut=30))
        set_if_none(kwargs, 'survival', ReferenceLineSurvival(ref_dirs))
        set_if_none(kwargs, 'eliminate_duplicates', True)

        super().__init__(**kwargs)

        self.func_display_attrs = disp_multi_objective

    def _solve(self, problem, termination):
        if self.ref_dirs.shape[1] != problem.n_obj:
            raise Exception(
                "Dimensionality of reference points must be equal to the number of objectives: %s != %s" %
                (self.ref_dirs.shape[1], problem.n_obj))

        return super()._solve(problem, termination)


def comp_by_cv_then_random(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop.CV[a, 0] > 0.0 or pop.CV[b, 0] > 0.0:
            S[i] = compare(a, pop.CV[a, 0], b, pop.CV[b, 0], method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible just set random
        else:
            S[i] = random.choice([a, b])

    return S[:, None].astype(np.int)



import numpy as np

from pymoo.util.dominator import Dominator


class NonDominatedSorting:

    def __init__(self, epsilon=0.0, method="cython_fast_non_dominated_sort") -> None:
        super().__init__()
        self.epsilon = float(epsilon)
        self.method = method

    def do(self, F, return_rank=False, only_non_dominated_front=False, n_stop_if_ranked=None):

        # if not set just set it to a very large values because the cython methods do not take None
        if n_stop_if_ranked is None:
            n_stop_if_ranked = int(1e8)

        if self.method == 'fast_non_dominated_rank':
            M = Dominator.calc_domination_matrix(F, epsilon=self.epsilon)
            from pymoo.util.non_dominated_sorting import fast_non_dominated_sort
            fronts = fast_non_dominated_sort(M)

        elif self.method == 'cython_fast_non_dominated_sort':
            from pymoo.cython.non_dominated_sorting import fast_non_dominated_sort
            fronts = fast_non_dominated_sort(F, self.epsilon, n_stop_if_ranked=n_stop_if_ranked)

        elif self.method == 'cython_best_order_sort':
            from pymoo.cython.non_dominated_sorting import best_order_sort
            fronts = best_order_sort(F, self.epsilon)

        elif self.method == 'cython_fast_best_order_sort':
            from pymoo.cython.non_dominated_sorting import fast_best_order_sort
            fronts = fast_best_order_sort(F, self.epsilon)

        else:
            raise Exception("Unknown non-dominated sorting method: %s" % self.method)

        # convert to numpy array for each front and filter by n_stop_if_ranked if desired
        _fronts = []
        n_ranked = 0
        for front in fronts:

            _fronts.append(np.array(front, dtype=np.int))

            # increment the n_ranked solution counter
            n_ranked += len(front)

            # stop if more than this solutions are n_ranked
            if n_ranked >= n_stop_if_ranked:
                break

        fronts = _fronts

        if only_non_dominated_front:
            return fronts[0]

        if return_rank:
            rank = rank_from_fronts(fronts, F.shape[0])
            return fronts, rank

        return fronts


def rank_from_fronts(fronts, n):
    # create the rank array and set values
    rank = np.full(n, 1e16, dtype=np.int)
    for i, front in enumerate(fronts):
        rank[front] = i

    return rank

# Returns all indices of F that are not dominated by the other objective values
def find_non_dominated(F, _F=None):
    M = Dominator.calc_domination_matrix(F, _F)
    I = np.where(np.all(M >= 0, axis=1))[0]
    return I


def non_dominated_sort_naive(M):
    fronts = []
    remaining = set(range(M.shape[0]))

    while len(remaining) > 0:

        front = []

        for i in remaining:

            is_dominated = False
            dominating = set()

            for j in front:
                rel = M[i, j]
                if rel == 1:
                    dominating.add(j)
                elif rel == -1:
                    is_dominated = True
                    break

            if is_dominated:
                continue
            else:
                front = [x for x in front if x not in dominating]
                front.append(i)

        [remaining.remove(e) for e in front]
        fronts.append(front)

    return fronts


def fast_non_dominated_sort(M):
    # calculate the dominance matrix
    n = M.shape[0]

    fronts = []

    if n == 0:
        return fronts

    # final rank that will be returned
    n_ranked = 0
    ranked = np.zeros(n, dtype=np.int)

    # for each individual a list of all individuals that are dominated by this one
    is_dominating = [[] for _ in range(n)]

    # storage for the number of solutions dominated this one
    n_dominated = np.zeros(n)

    current_front = []

    for i in range(n):

        for j in range(i + 1, n):
            rel = M[i, j]
            if rel == 1:
                is_dominating[i].append(j)
                n_dominated[j] += 1
            elif rel == -1:
                is_dominating[j].append(i)
                n_dominated[i] += 1

        if n_dominated[i] == 0:
            current_front.append(i)
            ranked[i] = 1.0
            n_ranked += 1

    # append the first front to the current front
    fronts.append(current_front)

    # while not all solutions are assigned to a pareto front
    while n_ranked < n:

        next_front = []

        # for each individual in the current front
        for i in current_front:

            # all solutions that are dominated by this individuals
            for j in is_dominating[i]:
                n_dominated[j] -= 1
                if n_dominated[j] == 0:
                    next_front.append(j)
                    ranked[j] = 1.0
                    n_ranked += 1

        fronts.append(next_front)
        current_front = next_front

    return fronts



import numpy as np

from pymoo.model.survival import Survival, split_by_feasibility


class FitnessSurvival(Survival):

    def _do(self, pop, n_survive, out=None, **kwargs):

        if pop.F.shape[1] != 1:
            raise ValueError("FitnessSurvival can only used for single objective problems!")

        # split by feasibility
        feasible, infeasible = split_by_feasibility(pop)
        survivors = np.array([], dtype=np.int)

        # if there are feasible solutions add them first
        if len(feasible) > 0:
            survivors = feasible[np.argsort(pop.F[feasible, 0])]

        # if we can select only from feasible - truncate here
        if len(feasible) > n_survive:
            survivors = survivors[:n_survive]

        # otherwise fill up with least infeasible ones
        else:
            n_infeasible = (n_survive - len(feasible))
            survivors = np.concatenate([survivors, infeasible[:n_infeasible]])

        pop.filter(survivors)



from pymoo.model.mutation import Mutation
from pymoo.rand import random


class BinaryBitflipMutation(Mutation):
    def __init__(self, p_mut=None):
        self.p_mut = p_mut

    def _do(self, p, X, Y, **kwargs):

        if self.p_mut is None:
            self.p_mut = 1.0 / len(X)

        for i in range(X.shape[0]):

            for j in range(X.shape[1]):

                if random.random() < self.p_mut:
                    Y[i,j] = not X[i,j]
                else:
                    Y[i, j] = X[i, j]

        return Y



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, cdivision=True


cdef double c_norm(double[:] v)
cdef double[:,:] c_calc_perpendicular_distance(double[:,:] P, double[:,:] L)


from pymoo.operators.crossover.bin_uniform_crossover import BinaryUniformCrossover
from pymoo.operators.crossover.real_simulated_binary_crossover import SimulatedBinaryCrossover
from pymoo.operators.mutation.bin_bitflip_mutation import BinaryBitflipMutation
from pymoo.operators.mutation.real_polynomial_mutation import PolynomialMutation
from pymoo.operators.sampling.bin_random_sampling import BinaryRandomSampling
from pymoo.operators.sampling.real_random_sampling import RealRandomSampling
from pymoo.operators.selection.random_selection import RandomSelection


def set_if_none(kwargs, str, val):
    if str not in kwargs:
        kwargs[str] = val


def set_default_if_none(var_type, kwargs):
    set_if_none(kwargs, 'pop_size', 100)
    set_if_none(kwargs, 'disp', False)
    set_if_none(kwargs, 'selection', RandomSelection())
    set_if_none(kwargs, 'survival', None)

    # values for mating
    if var_type == "real":
        set_if_none(kwargs, 'sampling', RealRandomSampling())
        set_if_none(kwargs, 'crossover', SimulatedBinaryCrossover(prob_cross=0.9, eta_cross=20))
        set_if_none(kwargs, 'mutation', PolynomialMutation(prob_mut=None, eta_mut=15))
    elif var_type == "binary":
        set_if_none(kwargs, 'sampling', BinaryRandomSampling())
        set_if_none(kwargs, 'crossover', BinaryUniformCrossover())
        set_if_none(kwargs, 'mutation', BinaryBitflipMutation())
        set_if_none(kwargs, 'eliminate_duplicates', True)



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, cdivision=True


from my_math cimport c_norm
from libcpp.vector cimport vector

import numpy as np


def cython_pbi(double[:,:] F, double[:,:] weights, double[:] ideal_point, double theta):
    return np.array(c_pbi(F, weights, ideal_point, theta), dtype=np.float)



cdef extern from "math.h":
    double sqrt(double m)
    double pow(double base, double exponent)


cdef vector[double] c_pbi(double[:,:] F, double[:,:] weights, double[:] ideal_point, double theta):
    cdef:
        double d1, d2, f_max, norm
        int i, j, n_dim
        vector[double] pbi

    n_points = F.shape[0]
    n_obj = F.shape[1]
    pbi = vector[double](n_points)


    for i in range(n_points):

        norm = c_norm(weights[i,:])

        d1 = 0
        for j in range(n_obj):
            d1 += pow(((F[i,j] - ideal_point[j]) * weights[i,j]) / norm, 2.0)
        d1 = sqrt(d1)

        d2 = 0
        for j in range(n_obj):
            d2 += pow(F[i,j] - (ideal_point[j] - d1 * weights[i,j]), 2.0)
        d2 = sqrt(d2)

        pbi[i] = d1 + theta * d2

    return pbi





import numpy as np

from pymoo.algorithms.nsga2 import NSGA2
from pymoo.algorithms.nsga3 import NSGA3
from pymoo.algorithms.rnsga3 import RNSGA3
from pymoo.algorithms.so_genetic_algorithm import SingleObjectiveGeneticAlgorithm
from pymoo.algorithms.unsga3 import UNSGA3
from pymoo.model.termination import MaximumFunctionCallTermination, MaximumGenerationTermination, IGDTermination
from pymoo.rand import random
from pymop.problem import Problem


def minimize(fun,
             method,
             xl=None,
             xu=None,
             termination=('n_gen', 200),
             n_var=None,
             fun_args={},
             method_args={},
             seed=None,
             callback=None,
             disp=True,
             save_history=False):

    """

    Minimization of function of one or more variables, objectives and constraints.

    This is used as a convenience function to execute several algorithms with default settings which turned
    out to work for a test problems. However, evolutionary computations utilizes the idea of customizing a
    meta-algorithm. Customizing the algorithm using the object oriented interface is recommended to improve the
    convergence.

    Parameters
    ----------

    fun : callable
        A function that gets X which is a 2d array as input. Each row represents a solution to evaluate
        and each column a variable. The function needs to return also the same number of rows and each column
        is one objective to optimize. In case of constraints, a second 2d array can be returned.
    xl : numpy.array or number
        The lower boundaries of variables as a 1d array
    xu : numpy.array or number
        The upper boundaries of variables as a 1d array
    n_var : int
        If xl or xu is only a number, this is used to create the boundary numpy arrays. Other it remains unused.
    termination : tuple
        The termination criterium that is used to stop the algorithm when the result is satisfying.
    fun_args : dict
        Additional arguments if necessary to evaluate the solutions (constants, random seed, ...)
    method : string
        Algorithm that is used to solve the problem.
    method_args : dict
        Additional arguments to initialize the algorithm object
    seed : int
        Random seed to be used for the run
    callback : callable, optional
        Called after each iteration, as ``callback(D)``, where ``D`` is a dictionary with
        algorithm information, such as the current design, objective and constraint space.
    disp : bool
        Whether to display each generation the current result or not.
    save_history : bool
        If true every iteration a snapshot of the algorithm is stored.

    Returns
    -------
    res : Result
        The optimization result represented as a ``Result`` object.


    """

    if isinstance(fun, Problem):
        problem = fun

    elif callable(fun):

        if xl is None or xu is None:
            raise Exception("Please provide lower and upper bounds for the problem.")
        if isinstance(xl, (int, float)):
            xl = xl * np.ones(n_var)
        if isinstance(xu, (int, float)):
            xu = xu * np.ones(n_var)

        # determine through a test evaluation details about the problem
        n_var = xl.shape[0]
        n_obj = -1
        n_constr = 0

        res = fun(xl[None, :], **fun_args)

        if isinstance(res, tuple):
            # if there are constraints it is a tuple of length two
            if len(res) > 1:
                n_constr = res[1].shape[1]
            n_obj = res[0].shape[1]
        else:
            n_obj = res.shape[1]

        # create an optimization problem given the function properties
        class OptimizationProblem(Problem):
            def __init__(self):
                Problem.__init__(self)
                self.n_var = n_var
                self.n_constr = n_constr
                self.n_obj = n_obj
                self.func = self._evaluate
                self.xl = xl
                self.xu = xu

            def _evaluate(self, x, f, g=None):
                if g is None:
                    f[:, :] = fun(x, **fun_args)
                else:
                    f[:, :], g[:, :] = fun(x, **fun_args)

        problem = OptimizationProblem()

    else:
        raise Exception("fun arguments needs to be either a function or an object of the class problem.")

    # create an evaluator defined by the termination criterium
    termination_criterium, termination_val = termination
    if termination_criterium == 'n_eval':
        termination = MaximumFunctionCallTermination(termination_val)
    elif termination_criterium == 'n_gen':
        termination = MaximumGenerationTermination(termination_val)
    elif termination_criterium == 'igd':
        termination = IGDTermination(problem, termination_val)
    else:
        raise Exception('Unknown Termination criterium: %s' % termination_criterium)

    # set a random random seed if not provided
    if seed is None:
        seed = random.randint(1, 10000)

    return _minimize(problem, termination, method, method_args=method_args, seed=seed,
                     callback=callback, disp=disp, save_history=save_history)


def _minimize(problem, termination, method, method_args={}, seed=1,
              callback=None, disp=False, save_history=False):
    """
        See :func:`~pymoo.optimize.minimize` for description. Instead of a function the parameter is a problem class.
    """

    # choose the algorithm implementation given the string
    if method == 'nsga2':
        algorithm = NSGA2(**method_args)
    elif method == 'nsga3':
        algorithm = NSGA3(**method_args)
    elif method == 'rnsga3':
        algorithm = RNSGA3(**method_args)
    elif method == 'unsga3':
        algorithm = UNSGA3(**method_args)
    elif method == 'ga':
        algorithm = SingleObjectiveGeneticAlgorithm(**method_args)
    else:
        raise Exception('Unknown method: %s' % method)

    res = algorithm.solve(problem,
                          termination,
                          disp=disp,
                          callback=callback,
                          seed=seed,
                          return_only_feasible=False,
                          return_only_non_dominated=False,
                          save_history=save_history)

    return res



import time

import numpy as np

from pymoo.util.plotting import plot, animate
from pymop.problems.zdt import ZDT1


def run():

    # create the optimization problem
    problem = ZDT1()

    start_time = time.time()

    # solve the given problem using an optimization algorithm (here: nsga2)
    from pymoo.optimize import minimize
    res = minimize(problem,
                   method='nsga2',
                   method_args={'pop_size': 100},
                   termination=('n_gen', 200),
                   seed=1,
                   save_history=True,
                   disp=True)
    F = res['F']

    print("--- %s seconds ---" % (time.time() - start_time))

    scatter_plot = True
    save_animation = False

    if scatter_plot:
        plot(F, problem)

    if save_animation:
        H = np.concatenate([e['pop'].F[None, :, :] for e in res['history']], axis=0)
        animate('%s.mp4' % problem.name(), H, problem)


if __name__ == '__main__':
    run()


import numpy as np

from pymoo.rand.random_generator import RandomGenerator


class NumpyRandomGenerator(RandomGenerator):

    def _seed(self, x):
        np.random.seed(x)

    def random(self, size=None):
        return np.random.random(size)

    def randint(self, low, high=None, size=None):
        return np.random.randint(low, high=high, size=size)

    def perm(self, n):
        return np.random.permutation(n)




import numpy as np

from pymoo.model.survival import Survival, split_by_feasibility
from pymoo.operators.survival.reference_line_survival import associate_to_niches, calc_niche_count, niching, \
    get_extreme_points, get_intercepts
from pymoo.util.mathematics import Mathematics
from pymoo.util.non_dominated_sorting import NonDominatedSorting


class AspirationPointSurvival(Survival):
    def __init__(self, ref_points, aspiration_ref_dirs, mu=0.1):
        super().__init__()

        self.aspiration_ref_dirs = aspiration_ref_dirs
        self.extreme_points = None
        self.intercepts = None
        self.ideal_point = np.full(aspiration_ref_dirs.shape[1], np.inf)

        self.ref_points = ref_points
        self.mu = mu

    def _do(self, pop, n_survive, D=None, **kwargs):

        # convert to integer for later usage
        n_survive = int(n_survive)

        # first split by feasibility for normalization
        feasible, infeasible = split_by_feasibility(pop)

        # number of survivors from the feasible population 
        # in case of having not enough feasible solution all feasible will survive
        if len(feasible) < n_survive:
            n_survive_feasible = len(feasible)
        else:
            n_survive_feasible = n_survive

        # attributes to be set after the survival
        survivors, rank, niche_of_individuals, dist_to_niche = [], [], [], []

        # if there are feasible solutions to survive
        if len(feasible) > 0:

            # consider only feasible solutions form now on
            F = pop.F[feasible, :]

            # find or usually update the new ideal point - from feasible solutions
            self.ideal_point = np.min(np.concatenate([self.ideal_point[None, :], F], axis=0), axis=0)

            # calculate the fronts of the population
            fronts, _rank = NonDominatedSorting(epsilon=Mathematics.EPS).do(F, return_rank=True,
                                                                            n_stop_if_ranked=n_survive_feasible)
            non_dominated, last_front = fronts[0], fronts[-1]

            # calculate the worst point of feasible individuals
            worst_point = np.max(np.vstack((F, self.ref_points)), axis=0)
            # calculate the nadir point from non dominated individuals
            nadir_point = np.max(np.vstack((F[non_dominated, :], self.ref_points)), axis=0)

            # find the extreme points for normalization
            self.extreme_points = get_extreme_points(np.vstack((F, self.ref_points)), self.ideal_point,
                                                     extreme_points=self.extreme_points)

            # find the intercepts for normalization and do backup if gaussian elimination fails
            self.intercepts = get_intercepts(self.extreme_points, self.ideal_point, nadir_point, worst_point)

            unit_ref_points = (self.ref_points - self.ideal_point) / self.intercepts
            ref_dirs = get_ref_dirs_from_points(unit_ref_points, self.aspiration_ref_dirs, mu=self.mu)

            # index of the first n fronts form now on - including splitting front
            I = np.concatenate(fronts)
            F = F[I, :]

            # associate individuals to niches
            niche_of_individuals, dist_to_niche = associate_to_niches(F, ref_dirs, self.ideal_point,
                                                                      self.intercepts)

            # if a splitting of the last front is not necessary
            if F.shape[0] == n_survive_feasible:
                _survivors = np.arange(F.shape[0])

            # otherwise we have to select using niching
            else:

                _last_front = np.arange(len(I) - len(last_front), len(I))
                _until_last_front = np.arange(0, len(I) - len(last_front))

                _survivors = []
                n_remaining = n_survive_feasible
                niche_count = np.zeros(len(ref_dirs), dtype=np.int)

                if len(fronts) > 1:
                    _survivors.extend(_until_last_front)
                    niche_count = calc_niche_count(len(ref_dirs),
                                                   niche_of_individuals[_until_last_front])
                    n_remaining -= len(_until_last_front)

                S = niching(F[_last_front, :], n_remaining, niche_count, niche_of_individuals[_last_front],
                            dist_to_niche[_last_front])

                # S = niching_vectorized(F[_last_front, :], n_remaining, niche_count, niche_of_individuals[_last_front],
                #                       dist_to_niche[_last_front])

                _survivors.extend(_last_front[S].tolist())

            # reindex the survivors to the absolute index
            survivors = feasible[I[_survivors]]

            # save the attributes for surviving individuals
            rank = _rank[I[_survivors]]
            niche_of_individuals = niche_of_individuals[_survivors]
            dist_to_niche = dist_to_niche[_survivors]

        # if we need to fill up with infeasible solutions - we do so. Also, the data structured need to be reindexed
        n_infeasible = n_survive - len(survivors)
        if n_infeasible > 0:
            survivors = np.concatenate([survivors, infeasible[:n_infeasible]])
            rank = np.concatenate([rank, Mathematics.INF * np.ones(n_infeasible)])
            niche_of_individuals = np.concatenate([niche_of_individuals, -1 * np.ones(n_infeasible)])
            dist_to_niche = np.concatenate([dist_to_niche, Mathematics.INF * np.ones(n_infeasible)])

        # set attributes globally for other modules
        if D is not None:
            D['rank'] = rank
            D['niche'] = niche_of_individuals
            D['dist_to_niche'] = dist_to_niche

        # now truncate the population
        pop.filter(survivors)


def get_ref_dirs_from_points(ref_point, ref_dirs, mu=0.1):
    """
    This function takes user specified reference points, and creates smaller sets of equidistant
    Das-Dennis points around the projection of user points on the Das-Dennis hyperplane
    :param ref_point: List of user specified reference points
    :param n_obj: Number of objectives to consider
    :param mu: Shrinkage factor (0-1), Smaller = tigher convergence, Larger= larger convergence
    :return: Set of reference points
    """

    n_obj = ref_point.shape[1]

    val = []
    n_vector = np.ones(n_obj) / np.sqrt(n_obj)  # Normal vector of Das Dennis plane
    point_on_plane = np.eye(n_obj)[0]  # Point on Das-Dennis

    for point in ref_point:

        ref_dir_for_aspiration_point = np.copy(ref_dirs)  # Copy of computed reference directions
        ref_dir_for_aspiration_point = mu * ref_dir_for_aspiration_point

        cent = np.mean(ref_dir_for_aspiration_point, axis=0)  # Find centroid of shrunken reference points

        # Project shrunken Das-Dennis points back onto original Das-Dennis hyperplane
        intercept = line_plane_intersection(np.zeros(n_obj), point, point_on_plane, n_vector)
        shift = intercept - cent  # shift vector

        ref_dir_for_aspiration_point += shift

        # If reference directions are located outside of first octant, redefine points onto the border
        if not (ref_dir_for_aspiration_point > 0).min():
            ref_dir_for_aspiration_point[ref_dir_for_aspiration_point < 0] = 0
            ref_dir_for_aspiration_point = ref_dir_for_aspiration_point / np.sum(ref_dir_for_aspiration_point, axis=1)[
                                                                          :, None]
        val.extend(ref_dir_for_aspiration_point)

    val.extend(np.eye(n_obj))  # Add extreme points
    return np.array(val)


# intersection function

def line_plane_intersection(l0, l1, p0, p_no, epsilon=1e-6):
    """
    l0, l1: define the line
    p0, p_no: define the plane:
        p0 is a point on the plane (plane coordinate).
        p_no is a normal vector defining the plane direction;
             (does not need to be normalized).

    reference: https://en.wikipedia.org/wiki/Line%E2%80%93plane_intersection
    return a Vector or None (when the intersection can't be found).
    """

    l = l1 - l0
    dot = np.dot(l, p_no)

    if abs(dot) > epsilon:
        # the factor of the point between p0 -> p1 (0 - 1)
        # if 'fac' is between (0 - 1) the point intersects with the segment.
        # otherwise:
        #  < 0.0: behind p0.
        #  > 1.0: infront of p1.
        w = p0 - l0
        d = np.dot(w, p_no) / dot
        l = l * d
        return l0 + l
    else:
        # The segment is parallel to plane then return the perpendicular projection
        ref_proj = l1 - (np.dot(l1 - p0, p_no) * p_no)
        return ref_proj



from pymoo.model.crossover import Crossover
from pymoo.rand import random


class BinaryUniformCrossover(Crossover):
    def __init__(self):
        super().__init__(2, 2)

    def _do(self, p, parents, children, **kwargs):

        # number of parents
        n_parents = parents.shape[0]

        # random matrix to do the crossover
        M = random.random(n_parents, p.n_var)
        smaller, larger = M < 0.5, M > 0.5

        # first possibility where first parent 0 is copied
        children[:n_parents][smaller] = parents[:, 0, :][smaller]
        children[:n_parents][larger] = parents[:, 1, :][larger]

        # now flip the order of parents with the same random array and write the second half of children
        children[n_parents:][smaller] = parents[:, 1, :][smaller]
        children[n_parents:][larger] = parents[:, 0, :][larger]



import copy

import numpy as np


class Population:

    def __init__(self, **kwargs):
        self.D = dict(kwargs)

    def __getattr__(self, name):

        # if internal function method just return the one whats usually done
        if str.startswith(name, '__'):
            return super().__getattr__(name)

        # if we really ask for the dictionary
        if name == "D":
            return self.__dict__.get("D", None)
        # else we ask for an entry in the dictionary
        else:
            if "D" in self.__dict__:
                return self.__dict__["D"].get(name, None)
            else:
                print(name)
                return {}

    def __setattr__(self, name, value):
        if name == 'D':
            self.__dict__['D'] = value
        else:
            self.__dict__['D'][name] = value

    def __deepcopy__(self, a):
        return Population(**copy.deepcopy(self.D))

    def merge(self, other):
        D = {}
        for key, value in self.D.items():

            # key must be in both populations
            if key not in other.D:
                continue

            if isinstance(value, np.ndarray):
                try:
                    D[key] = np.concatenate([self.D[key], other.D[key]])
                except:
                    D[key] = self.D[key]

        # all values in other that were not present in self
        for key, value in other.D.items():
            if key not in self.D:
                D[key] = other.D[key]

        self.D = D

    def size(self):
        if self.X is not None:
            return self.X.shape[0]
        elif self.F is not None:
            return self.F.shape[0]
        else:
            return None

    def copy(self):
        pop = Population()
        pop.D = copy.deepcopy(self.D)
        return pop

    def filter(self, v):

        if isinstance(v, np.ndarray):
            v = v.astype(np.int)

        D = {}
        for key, value in self.D.items():
            if value is not None and isinstance(value, np.ndarray):
                D[key] = value[v, :]
            else:
                D[key] = value
        self.D = D



import numpy as np

from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.operators.crossover.real_simulated_binary_crossover import SimulatedBinaryCrossover
from pymoo.operators.default_operators import set_if_none
from pymoo.operators.mutation.real_polynomial_mutation import PolynomialMutation
from pymoo.operators.sampling.real_random_sampling import RealRandomSampling
from pymoo.operators.selection.tournament_selection import TournamentSelection
from pymoo.operators.survival.rank_and_crowding import RankAndCrowdingSurvival
from pymoo.util.display import disp_multi_objective
from pymoo.util.dominator import Dominator, compare


class NSGA2(GeneticAlgorithm):

    def __init__(self, pop_size=100, **kwargs):

        # default settings for nsga2 - not overwritten if provided as kwargs
        set_if_none(kwargs, 'pop_size', pop_size)
        set_if_none(kwargs, 'sampling', RealRandomSampling())
        set_if_none(kwargs, 'selection', TournamentSelection(func_comp=comp_by_dom_and_crowding))
        set_if_none(kwargs, 'crossover', SimulatedBinaryCrossover(prob_cross=0.9, eta_cross=15))
        set_if_none(kwargs, 'mutation', PolynomialMutation(prob_mut=None, eta_mut=20))
        set_if_none(kwargs, 'survival', RankAndCrowdingSurvival())
        set_if_none(kwargs, 'eliminate_duplicates', True)

        super().__init__(**kwargs)

        self.D['tournament_type'] = 'comp_by_dom_and_crowding'
        self.func_display_attrs = disp_multi_objective



def comp_by_dom_and_crowding(pop, P, D, **kwargs):
    if P.shape[1] != 2:
        raise ValueError("Only implemented for binary tournament!")

    rank, crowding, tournament_type = D['rank'], D['crowding'], D['tournament_type']

    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):

        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop.CV[a, 0] > 0.0 or pop.CV[b, 0] > 0.0:
            S[i] = compare(a, pop.CV[a, 0], b, pop.CV[b, 0], method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible
        else:

            if tournament_type == 'comp_by_dom_and_crowding':
                rel = Dominator.get_relation(pop.F[a, :], pop.F[b, :])
                if rel == 1:
                    S[i] = a
                elif rel == -1:
                    S[i] = b

            elif tournament_type == 'comp_by_rank_and_crowding':
                S[i] = compare(a, rank[a], b, rank[b], method='smaller_is_better')

            else:
                raise Exception("Unknown tournament type.")

            # if rank or domination relation didn't make a decision compare by crowding
            if np.isnan(S[i]):
                S[i] = compare(a, crowding[a], b, crowding[b], method='larger_is_better',
                               return_random_if_equal=True)

    return S[:, None].astype(np.int)



from pymoo.model.sampling import Sampling
from pymoo.rand import random


class RealRandomSampling(Sampling):
    """
    Randomly sample points in the real space by considering the lower and upper bounds of the problem.
    """

    def sample(self, problem, n_samples, **kwargs):
        m = len(problem.xl)
        val = random.random(size=(n_samples, m))
        for i in range(m):
            val[:, i] = val[:, i] * (problem.xu[i] - problem.xl[i]) + problem.xl[i]
        return val



import numpy as np
from scipy.spatial.distance import cdist

from pymoo.model.indicator import Indicator


class GD(Indicator):
    def __init__(self, true_front):
        Indicator.__init__(self)
        self.pareto_front = true_front

    def _calc(self, F):
        v = cdist(F, self.pareto_front)
        return np.mean(np.min(v, axis=1))



import numpy as np

from pymoo.indicators.gd import GD
from pymoo.indicators.igd import IGD


def disp_single_objective(problem, evaluator, D):
    attrs = [('n_gen', D['n_gen'], 5),
             ('n_eval', evaluator.n_eval, 7)]

    if problem.n_constr > 0:
        attrs.append(('cv (min/avg)', "%.5f / %.5f" % (np.min(D['pop'].CV), np.mean(D['pop'].CV)), 13))

    attrs.extend([
             ('favg', "%.5f" % np.mean(D['pop'].F), 5),
             ('fopt', "%.10f" % np.min(D['pop'].F), 5)
             ])

    return attrs


def disp_multi_objective(problem, evaluator, D):
    attrs = [('n_gen', D['n_gen'], 5),
             ('n_eval', evaluator.n_eval, 7)]

    if problem.n_constr > 0:
        attrs.append(('cv (min/avg)', "%.5f / %.5f" % (np.min(D['pop'].CV), np.mean(D['pop'].CV)), 13))

    pf = problem.pareto_front()
    if pf is not None:
        attrs.append(('igd', "%.5f" % IGD(pf).calc(D['pop'].F), 8))
        attrs.append(('gd', "%.5f" % GD(pf).calc(D['pop'].F), 8))

    return attrs



from pymoo.indicators.igd import IGD


class Termination:

    def do_continue(self, D):
        return not self.has_finished(D)

    def has_finished(self, D):
        return not self.do_continue(D)


class MaximumFunctionCallTermination(Termination):

    def __init__(self, n_max_evals) -> None:
        super().__init__()
        self.n_max_evals = n_max_evals

    def do_continue(self, D):
        return D['evaluator'].n_eval < self.n_max_evals


class MaximumGenerationTermination(Termination):

    def __init__(self, n_max_gen) -> None:
        super().__init__()
        self.n_max_gen = n_max_gen

    def do_continue(self, D):
        return D['n_gen'] < self.n_max_gen


class IGDTermination(Termination):

    def __init__(self, problem, igd) -> None:
        super().__init__()
        pf = problem.pareto_front()

        if pf is None:
            raise Exception("You can only use IGD termination criteria if the pareto front is known!")

        self.obj = IGD(pf)
        self.igd = igd

    def do_continue(self, D):
        return self.obj.calc(D['pop'].F) > self.igd



from abc import abstractmethod

import numpy as np


class Crossover:
    """
    The crossover combines parents to offsprings. Some crossover are problem specific and use additional information.
    This class must be inherited from to provide a crossover method to an algorithm.
    """

    def __init__(self, n_parents, n_children):
        self.n_parents = n_parents
        self.n_children = n_children

    def do(self, problem, parents, **kwargs):
        """

        This method executes the crossover on the parents. This class wraps the implementation of the class
        that implements the crossover.

        Parameters
        ----------
        problem: class
            The problem to be solved. Provides information such as lower and upper bounds or feasibility
            conditions for custom crossovers.

        parents: numpy.ndarray
            The parents as a matrix. Each row is a parent and the columns are equal to the parameter length.

        kwargs : dict:
            Any additional data that might be necessary to perform the crossover. E.g. constants of an algorithm.

        Returns
        -------
        children: np.ndarray
            The children as a matrix. n_children rows and the number of columns is equal to the variable
            length of the problem.

        """
        n_matings = parents.shape[1]
        n_var = parents.shape[2]

        if n_var != problem.n_var:
            raise ValueError('Exception during crossover: Variable length is not equal to the defined one in problem.')

        if self.n_parents != parents.shape[0]:
            raise ValueError('Exception during crossover: Number of parents differs from defined at crossover.')

        children = np.full((n_matings * self.n_children, n_var), np.inf)
        self._do(problem, parents, children, **kwargs)
        return children

    @abstractmethod
    def _do(self, problem, parents, children, **kwargs):
        n_children = 0
        for k in range(parents.shape[1]):

            _children = np.full((self.n_children, problem.n_var), np.inf)

            self._mating(problem, parents[:, k, :], _children, **kwargs)

            for i in range(_children.shape[0]):
                children[n_children + i, :] = _children[i, :]

            n_children += self.n_children

    @abstractmethod
    def _mating(self, problem, parents, children, **kwargs):
        pass



from abc import abstractmethod

import numpy as np


class RandomGenerator:
    """
    Implementation of a random generator used for all algorithm. This is just the base class which needs to
    be inherited from.
    """

    def seed(self, n):
        if n < 0:
            raise Exception("Random seed must be larger than 0!")
        self._seed(n)

    def perm(self, size):
        return np.argsort(self.random(size), kind='quicksort')

    def randint(self, low, high=None, size=None):
        val = self.random(size)
        return (low + (val * (high - low))).astype(np.int)

    def choice(self, a):
        return a[self.randint(0, len(a))]

    def random(self, size=None):
        if size is None:
            return self._rand((1, 1))[0, 0]

        elif isinstance(size, int):
            return self._rand((1, size))[0, :]

        elif isinstance(size, tuple):
            return self._rand(size)
        else:
            raise Exception("rand method not defined for: %s" % size)

    def _rand(self, size=None):
        val = np.full(size, np.inf)
        for index, _ in np.ndenumerate(val):
            val[index] = self._rand_float()
        return val

    @abstractmethod
    def _rand_float(self):
        pass

    @abstractmethod
    def _seed(self, s):
        pass



import numpy as np

from pymoo.configuration import Configuration


# use a singleton class to set the random generator globally
class Singleton:
    __instance = None

    @staticmethod
    def get_instance():
        if Singleton.__instance is None:
            Singleton.__instance = Configuration.rand
        return Singleton.__instance


def seed(s):
    """
    Set the random seed of the current random instance that is used.

    Parameters
    ----------
    s : int
        seed

    """
    Singleton.get_instance().seed(s)


def perm(n):
    """
    Returns a random permutation with length n.

    Parameters
    ----------
    n : int
        Length of the permutation array.

    Returns
    -------
    p : numpy.array
        Permutation array with type integer.

    """
    return Singleton.get_instance().perm(n).astype(np.int)


def random(size=None):
    """
    If size is None returns one random float [0.0,1.0), other an numpy array with the predefined size.

    Parameters
    ----------
    size : tuple
        (Optional): Shape of the numpy array

    """
    return Singleton.get_instance().random(size=size)


def randint(low, high=None, size=None):
    """
    Return random integers from low (inclusive) to high (exclusive).

    Return random integers from the “discrete uniform” distribution in the “half-open” interval [low, high).
    If high is None (the default), then results are from [0, low).

    """
    if high is None:
        return Singleton.get_instance().randint(0, high=low, size=size)
    else:
        return Singleton.get_instance().randint(low, high=high, size=size)


def choice(a):
    """
    Select randomly an element from a list.
    """
    return Singleton.get_instance().choice(a)



from abc import abstractmethod

import numpy as np


class Mutation:
    def do(self, problem, X, **kwargs):
        """

        Mutate variables in a genetic way.

        Parameters
        ----------
        problem : class
            The problem instance - specific information such as variable bounds might be needed.
        X : np.ndarray
            Two dimensional array where each row is an individual and each column represents a variable.

        Returns
        -------
        Y : np.ndarray
            The mutated individuals

        """
        Y = np.full(X.shape, np.inf)
        self._do(problem, X, Y, **kwargs)
        return Y

    @abstractmethod
    def _do(self, problem, X, **kwargs):
        pass



import numpy as np
from numpy.linalg import LinAlgError

from pymoo.cython.my_math import cython_calc_perpendicular_distance
from pymoo.model.survival import Survival, split_by_feasibility
from pymoo.rand import random
from pymoo.util.mathematics import Mathematics
from pymoo.util.non_dominated_sorting import NonDominatedSorting


class ReferenceLineSurvival(Survival):
    def __init__(self, ref_dirs):
        super().__init__()
        self.ref_dirs = ref_dirs
        self.extreme_points = None
        self.intercepts = None
        self.ideal_point = np.full(ref_dirs.shape[1], np.inf)

    def _do(self, pop, n_survive, D=None, **kwargs):

        # convert to integer for later usage
        n_survive = int(n_survive)

        # first split by feasibility for normalization
        feasible, infeasible = split_by_feasibility(pop)

        # number of survivors from the feasible population 
        # in case of having not enough feasible solution all feasible will survive
        if len(feasible) < n_survive:
            n_survive_feasible = len(feasible)
        else:
            n_survive_feasible = n_survive

        # attributes to be set after the survival
        survivors, rank, niche_of_individuals, dist_to_niche = [], [], [], []

        # if there are feasible solutions to survive
        if len(feasible) > 0:

            # consider only feasible solutions form now on
            F = pop.F[feasible, :]

            # find or usually update the new ideal point - from feasible solutions
            self.ideal_point = np.min(np.concatenate([self.ideal_point[None, :], F], axis=0), axis=0)

            # calculate the fronts of the population
            fronts, _rank = NonDominatedSorting(epsilon=Mathematics.EPS).do(F, return_rank=True,
                                                                            n_stop_if_ranked=n_survive_feasible)
            non_dominated, last_front = fronts[0], fronts[-1]

            # calculate the worst point of feasible individuals
            worst_point = np.max(F, axis=0)
            # calculate the nadir point from non dominated individuals
            nadir_point = np.max(F[non_dominated, :], axis=0)

            # find the extreme points for normalization
            self.extreme_points = get_extreme_points(F, self.ideal_point, extreme_points=self.extreme_points)

            # find the intercepts for normalization and do backup if gaussian elimination fails
            self.intercepts = get_intercepts(self.extreme_points, self.ideal_point, nadir_point, worst_point)

            # index of the first n fronts form now on - including splitting front
            I = np.concatenate(fronts)
            F = F[I, :]

            # associate individuals to niches
            niche_of_individuals, dist_to_niche = associate_to_niches(F, self.ref_dirs, self.ideal_point,
                                                                      self.intercepts)

            # if a splitting of the last front is not necessary
            if F.shape[0] == n_survive_feasible:
                _survivors = np.arange(F.shape[0])

            # otherwise we have to select using niching
            else:

                _last_front = np.arange(len(I) - len(last_front), len(I))
                _until_last_front = np.arange(0, len(I) - len(last_front))

                _survivors = []
                n_remaining = n_survive_feasible
                niche_count = np.zeros(len(self.ref_dirs), dtype=np.int)

                if len(fronts) > 1:
                    _survivors.extend(_until_last_front)
                    niche_count = calc_niche_count(len(self.ref_dirs), niche_of_individuals[_until_last_front])
                    n_remaining -= len(_until_last_front)

                S = niching(F[_last_front, :], n_remaining, niche_count, niche_of_individuals[_last_front],
                            dist_to_niche[_last_front])

                # S = niching_vectorized(F[_last_front, :], n_remaining, niche_count, niche_of_individuals[_last_front],
                #                       dist_to_niche[_last_front])

                _survivors.extend(_last_front[S].tolist())

            # reindex the survivors to the absolute index
            survivors = feasible[I[_survivors]]

            # save the attributes for surviving individuals
            rank = _rank[I[_survivors]]
            niche_of_individuals = niche_of_individuals[_survivors]
            dist_to_niche = dist_to_niche[_survivors]

        # if we need to fill up with infeasible solutions - we do so. Also, the data structured need to be reindexed
        n_infeasible = n_survive - len(survivors)
        if n_infeasible > 0:
            survivors = np.concatenate([survivors, infeasible[:n_infeasible]])
            rank = np.concatenate([rank, Mathematics.INF * np.ones(n_infeasible)])
            niche_of_individuals = np.concatenate([niche_of_individuals, -1 * np.ones(n_infeasible)])
            dist_to_niche = np.concatenate([dist_to_niche, Mathematics.INF * np.ones(n_infeasible)])

        # set attributes globally for other modules
        if D is not None:
            D['rank'] = rank
            D['niche'] = niche_of_individuals
            D['dist_to_niche'] = dist_to_niche

        # now truncate the population
        pop.filter(survivors)


def niching_vectorized(F, n_survive, niche_count, niche_of_individuals, dist_to_niche):
    survivors = []
    n_niches = len(niche_count)

    # maximum niche count of the last front
    max_niche_count = np.max(np.unique(niche_of_individuals, return_counts=True)[1])

    # stores for each niche the corresponding individuals - (-1=does not exist, -2=added before, i=otherwise)
    individuals_of_niches = np.full((n_niches, max_niche_count + 1), -1, dtype=np.int)

    # the minimum distance to the niche - just used to find the closest niche
    min_dist_closest_to_niche = np.full(n_niches, np.inf)

    # closest individual to niche - (NOTE: it is just an index to individuals_of_niches!)
    ind_closest_to_niche = np.full(n_niches, -1, dtype=np.int)

    # fill individuals_of_niches and ind_closest_to_niche with values
    counter = np.zeros(n_niches, dtype=np.int)
    for i in range(F.shape[0]):
        niche = niche_of_individuals[i]
        individuals_of_niches[niche, counter[niche]] = i

        if dist_to_niche[i] < min_dist_closest_to_niche[niche]:
            min_dist_closest_to_niche[niche] = dist_to_niche[i]
            ind_closest_to_niche[niche] = counter[niche]

        counter[niche] += 1

    # this mask is used to selected for the corresponding niches the next individuals
    mask = np.zeros(n_niches, dtype=np.int)

    # while we do have individuals to assign
    while len(survivors) < n_survive:

        # only consider niches that do have at least one individual assigned
        niches_remaining = np.where(individuals_of_niches[np.arange(n_niches), mask] != -1)[0]
        niche_count_remaining = niche_count[niches_remaining]

        # the minimum niche count of the remaining niches
        min_niche_count = np.min(niche_count_remaining)

        # get all niches with the minimum assignment
        selected_niches = niches_remaining[niche_count_remaining == min_niche_count]

        # remaining individuals which is the same as niches to select
        n_remaining = n_survive - len(survivors)

        # if only a few niches need to be select to fill up until n_survive - select the randomly
        if n_remaining < len(selected_niches):
            P = random.perm(len(selected_niches))
            selected_niches = selected_niches[P[:n_remaining]]

        # if we did not assign any individual
        if min_niche_count == 0:

            # select the closest individual to niche
            closest_individuals_of_niche = ind_closest_to_niche[selected_niches]
            S = individuals_of_niches[selected_niches, closest_individuals_of_niche]

            # since increasing does not work - here the selected are marked and skipped later
            individuals_of_niches[selected_niches, closest_individuals_of_niche] = -2

        else:
            # otherwise we just select one randomly through mask (values or not sorted)
            S = individuals_of_niches[selected_niches, mask[selected_niches]]
            # increase the mask by one for selected
            mask[selected_niches] += 1

        # if the mask points to an individual select by closest distance just point to the next
        mask[individuals_of_niches[np.arange(n_niches), mask] == -2] += 1

        # increase the niche count of the corresponding niche
        niche_count[selected_niches] += 1

        # save the individuals that survived
        survivors.extend(S)

    return survivors


def niching(F, n_remaining, niche_count, niche_of_individuals, dist_to_niche):
    survivors = []

    # boolean array of elements that are considered for each iteration
    mask = np.full(F.shape[0], True)

    while len(survivors) < n_remaining:

        # all niches where new individuals can be assigned to
        next_niches_list = np.unique(niche_of_individuals[mask])

        # pick a niche with minimum assigned individuals - break tie if necessary
        next_niche_count = niche_count[next_niches_list]
        next_niche = np.where(next_niche_count == next_niche_count.min())[0]
        next_niche = next_niches_list[next_niche]
        next_niche = next_niche[random.randint(0, len(next_niche))]

        # indices of individuals that are considered and assign to next_niche
        next_ind = np.where(np.logical_and(niche_of_individuals == next_niche, mask))[0]

        if len(next_ind) == 1:
            next_ind = next_ind[0]
        elif niche_count[next_niche] == 0:
            next_ind = next_ind[np.argmin(dist_to_niche[next_ind])]
        else:
            # not sorted so randomly the first is fine here
            next_ind = next_ind[0]
            # next_ind = next_ind[random.randint(0, len(next_ind))]

        mask[next_ind] = False
        survivors.append(int(next_ind))

        niche_count[next_niche] += 1

    return survivors


def get_extreme_points(F, ideal_point, extreme_points=None):
    # calculate the asf which is used for the extreme point decomposition
    asf = np.eye(F.shape[1])
    asf[asf == 0] = 1e-6

    # add the old extreme points to never loose them for normalization
    _F = F
    if extreme_points is not None:
        _F = np.concatenate([_F, extreme_points], axis=0)

    # update the extreme points for the normalization having the highest asf value each
    F_asf = np.max((_F - ideal_point) / asf[:, None, :], axis=2)
    extreme_points = _F[np.argmin(F_asf, axis=1), :]

    return extreme_points


def get_intercepts(extreme_points, ideal_point, nadir_point, worst_point):
    # normalization of the points in the new space
    nadir_point -= ideal_point
    worst_point -= ideal_point

    use_nadir = False

    try:
        # find the intercepts using gaussian elimination
        plane = np.linalg.solve(extreme_points - ideal_point, np.ones(extreme_points.shape[1]))

        # if the plane
        if np.any(plane == 0):
            use_nadir = True
        else:
            intercepts = 1 / plane

    except LinAlgError:
        use_nadir = True

    if use_nadir:
        intercepts = nadir_point

    # if even that point is too small
    if np.any(intercepts < 1e-6):
        intercepts = worst_point

    # if also the worst point is very small we set it to a small value, to avoid division by zero
    intercepts[intercepts < 1e-16] = 1e-16

    return intercepts


def associate_to_niches(F, niches, ideal_point, intercepts, utopianEpsilon=-0.001):
    # normalize by ideal point and intercepts
    N = (F - ideal_point) / intercepts

    # make sure that no values are 0. (subtracting a negative epsilon)
    N -= utopianEpsilon

    # dist_matrix = calc_perpendicular_dist_matrix(N, niches)
    dist_matrix = cython_calc_perpendicular_distance(N, niches)

    niche_of_individuals = np.argmin(dist_matrix, axis=1)
    dist_to_niche = dist_matrix[np.arange(F.shape[0]), niche_of_individuals]

    return niche_of_individuals, dist_to_niche


def calc_perpendicular_dist_matrix(N, ref_dirs):
    u = np.tile(ref_dirs, (len(N), 1))
    v = np.repeat(N, len(ref_dirs), axis=0)

    norm_u = np.linalg.norm(u, axis=1)

    scalar_proj = np.sum(v * u, axis=1) / norm_u
    proj = scalar_proj[:, None] * u / norm_u[:, None]
    val = np.linalg.norm(proj - v, axis=1)
    matrix = np.reshape(val, (len(N), len(ref_dirs)))

    return matrix


def calc_niche_count(n_niches, niche_of_individuals):
    niche_count = np.zeros(n_niches, dtype=np.int)
    index, count = np.unique(niche_of_individuals, return_counts=True)
    niche_count[index] = count
    return niche_count



from abc import abstractmethod

import numpy as np


class Survival:
    """
    The survival process is implemented inheriting from this class, which selects from a population only
    specific individuals to survive.
    """

    def __init__(self) -> None:
        super().__init__()

    def do(self, pop, n_survive, **kwargs):
        self._do(pop, n_survive, **kwargs)

    @abstractmethod
    def _do(self, pop, n_survive, **kwargs):
        pass


def split_by_feasibility(pop, sort_infeasbible_by_cv=True):
    # if no constraint violation is provided
    if pop.CV is None:
        return np.arange(pop.size()), np.array([])

    feasible, infeasible = [], []

    for i in range(pop.size()):
        if pop.CV[i, 0] <= 0:
            feasible.append(i)
        else:
            infeasible.append(i)

    if sort_infeasbible_by_cv:
        infeasible = sorted(infeasible, key=lambda i: pop.CV[i,:])

    return np.array(feasible), np.array(infeasible)



import numpy
import setuptools
from Cython.Build import cythonize
from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

__name__ = "pymoo"
__author__ = "Julian Blank"
__version__ = '0.2.1'
__url__ = "https://github.com/msu-coinlab/pymoo"



setup(
    name=__name__,
    version=__version__,
    author=__author__,
    url=__url__,
    python_requires='>3.3.0',
    author_email="blankjul@egr.msu.edu",
    description="Multi-Objective Optimization Algorithms",
    long_description=readme(),
    license='Apache License 2.0',
    keywords="optimization",
    packages=setuptools.find_packages(exclude=['tests', 'docs']),
    setup_requires=['cython'],
    install_requires=['pymop==0.2.1', 'numpy', 'scipy', 'matplotlib'],
    ext_modules=cythonize(
        "pymoo/cython/*.pyx",
        language="c++"
    ),
    include_dirs=[numpy.get_include()],
    include_package_data=True,
    platforms='any',
)



import numpy as np

from pymoo.model.survival import Survival, split_by_feasibility
from pymoo.util.mathematics import Mathematics
from pymoo.util.non_dominated_sorting import NonDominatedSorting
from pymoo.util.randomized_argsort import randomized_argsort


class RankAndCrowdingSurvival(Survival):
    def _do(self, pop, n_survive, D=None, **kwargs):

        # split by feasibility
        feasible, infeasible = split_by_feasibility(pop)

        # final result that contains indices, rank and crowding of surviving individuals
        survivors = []
        crowding = []
        rank = []

        if len(feasible) > 0:

            # calculate rank only of feasible solutions
            F = pop.F[feasible, :]
            fronts = NonDominatedSorting().do(F, n_stop_if_ranked=n_survive)

            # go through all fronts except the last one
            for k, front in enumerate(fronts):

                # calculate the crowding distance of the front
                #crowding_of_front = calc_crowding_distance(F[front, :])
                crowding_of_front = calc_crowding_distance(F[front, :])

                # current front sorted by crowding distance if splitting
                if len(survivors) + len(front) > n_survive:
                    I = randomized_argsort(crowding_of_front, order='descending', method='numpy')
                    I = I[:(n_survive - len(survivors))]

                # otherwise take the whole front
                else:
                    I = np.arange(len(front))

                # calculate crowding distance for the current front
                crowding.append(crowding_of_front[I])
                rank.append(np.array([k] * len(I)))
                survivors.extend(front[I])

            # create numpy arrays out of the lists
            rank = np.concatenate(rank)
            crowding = np.concatenate(crowding)

            # get absolute index from filtering before
            survivors = feasible[survivors]

        # if infeasible solutions need to be added - individuals sorted by constraint violation are added
        n_infeasible = (n_survive - len(survivors))
        if n_infeasible > 0:
            survivors = np.concatenate([survivors, infeasible[:n_infeasible]])
            rank = np.concatenate([rank, Mathematics.INF * np.ones(n_infeasible)])
            crowding = np.concatenate([crowding, -1.0 * np.ones(n_infeasible)])

        # now truncate the population
        pop.filter(survivors)

        if D is not None:
            D['rank'] = rank
            D['crowding'] = crowding


def calc_crowding_distance(F):
    infinity = 1e+14

    n_points = F.shape[0]
    n_obj = F.shape[1]

    if n_points <= 2:
        return np.full(n_points, infinity)
    else:

        # the final crowding distance result
        crowding = np.zeros(n_points)

        # sort each column and get index
        I = np.argsort(F, axis=0, kind='mergesort')

        # now really sort the whole array
        F = F[I, np.arange(n_obj)]

        # get the distance to the last element in sorted list and replace zeros with actual values
        dist = np.concatenate([F, np.full((1, n_obj), np.inf)]) \
               - np.concatenate([np.full((1, n_obj), -np.inf), F])

        index_dist_is_zero = np.where(dist == 0)

        dist_to_last = np.copy(dist)
        for i, j in zip(*index_dist_is_zero):
            dist_to_last[i, j] = dist_to_last[i - 1, j]

        dist_to_next = np.copy(dist)
        for i, j in reversed(list(zip(*index_dist_is_zero))):
            dist_to_next[i, j] = dist_to_next[i + 1, j]

        # normalize all the distances
        norm = np.max(F, axis=0) - np.min(F, axis=0)
        norm[norm == 0] = np.nan
        dist_to_last, dist_to_next = dist_to_last[:-1] / norm, dist_to_next[1:] / norm

        # if we divided by zero because all values in one columns are equal replace by none
        dist_to_last[np.isnan(dist_to_last)] = 0.0
        dist_to_next[np.isnan(dist_to_next)] = 0.0

        # sum up the distance to next and last and norm by objectives - also reorder from sorted list
        J = np.argsort(I, axis=0)
        crowding = np.sum(dist_to_last[J, np.arange(n_obj)] + dist_to_next[J, np.arange(n_obj)], axis=1) / n_obj

    # replace infinity with a large number
    crowding[np.isinf(crowding)] = infinity

    return crowding



if __name__ == '__main__':
    F = np.random.rand(100, 2)
    F[0] = np.max(F, axis=0)
    F[5] = F[0]
    F[50] = F[51]

    crowding = calc_crowding_distance(F)
    print(crowding[50], crowding[51])
    print(crowding[0], crowding[5])



import numpy as np

from pymoo.util.mathematics import Mathematics


class Decomposition:

    def do(self, F, **kwargs):
        return self._do(F, **kwargs)


class PenaltyBasedBoundaryInterception(Decomposition):

    def __init__(self, theta) -> None:
        super().__init__()
        self.theta = theta

    def _do(self, F, weights, ideal_point, **kwargs):

        n_points, n_weights = F.shape[0], weights.shape[0]

        if n_points == n_weights:
            pass
        elif n_points == 1 and n_weights > 1:
            F = np.repeat(F, n_weights, axis=0)
        elif n_points > 1 and n_weights == 1:
            weights = np.repeat(weights, n_points, axis=0)
        else:
            raise Exception("Either for each point a weight, one weight, or one objective value.")

        return self.python_pbi(F, weights, ideal_point, self.theta)
        # return cython_pbi(F, weights, ideal_point, self.theta)

    def python_pbi(self, F, weights, ideal_point, theta):
        d1 = np.linalg.norm((F - ideal_point) * weights, axis=1) / np.linalg.norm(weights, axis=1)
        d2 = np.linalg.norm(F - (ideal_point - d1[:, None] * weights), axis=1)
        return d1 + theta * d2

    def python_obi_jmetal(self, F, weights, ideal_point, theta):

        """
        double
        d1, d2, nl;
        double
        theta = 5.0;

        d1 = d2 = nl = 0.0;

        for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
        d1 += (individual.getObjective(i) - idealPoint[i]) * lambda[i];
        nl += Math.pow( lambda[i], 2.0);
        }
        nl = Math.sqrt(nl);
        d1 = Math.abs(d1) / nl;

        for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
        d2 += Math.pow((individual.getObjective(i) - idealPoint[i]) - d1 * ( lambda[i] / nl), 2.0);
        }
        d2 = Math.sqrt(d2);

        fitness = (d1 + theta * d2);
        """

        nl = weights / np.linalg.norm(weights, axis=1)[:, None]



        d1 = np.linalg.norm((F - ideal_point) * weights, axis=1)
        d2 = np.linalg.norm(F - (ideal_point - d1[:, None] * weights), axis=1)
        return d1 + theta * d2


class WeightedSum(Decomposition):

    def _do(self, F, weights, **kwargs):
        return np.sum(F * weights, axis=1)


class Tchebicheff(Decomposition):

    def _do(self, F, weights, ideal_point, **kwargs):
        v = np.abs((F - ideal_point - Mathematics.EPS) * weights)
        return np.max(v, axis=1)


class DoNotKnow(Decomposition):
    def _do(self, F, weights, ideal_point, **kwargs):
        alpha = 0.1
        v = (F - ideal_point) * (1 / F.shape[1] + alpha * weights)
        return np.max(v, axis=1)



from pymoo.rand import random


class BinaryRandomSampling:
    """
    Randomly sample a binary representation of 0's and 1's.
    """

    def sample(self, problem, n_samples, **kwargs):
        return random.randint(0, 2, size=(n_samples, problem.n_var))



import scipy

import numpy as np


def swap(M, a, b):
    tmp = M[a]
    M[a] = M[b]
    M[b] = tmp


# repairs a numpy array to be in bounds
def repair(X, xl, xu):
    larger_than_xu = X[0, :] > xu
    X[0, larger_than_xu] = xu[larger_than_xu]

    smaller_than_xl = X[0, :] < xl
    X[0, smaller_than_xl] = xl[smaller_than_xl]

    return X


def unique_rows(a):
    a = np.ascontiguousarray(a)
    unique_a = np.unique(a.view([('', a.dtype)] * a.shape[1]))
    return unique_a.view(a.dtype).reshape((unique_a.shape[0], a.shape[1]))


def parameter_less_constraints(F, CV, F_max=None):
    if F_max is None:
        F_max = np.max(F)
    has_constraint_violation = CV > 0
    F[has_constraint_violation] = CV[has_constraint_violation] + F_max
    return F


def random_permuations(n, l):
    from pymoo.rand import random
    perms = []
    for i in range(n):
        perms.append(random.perm(l))
    P = np.concatenate(perms)
    return P


def get_duplicates(M):
    res = []
    I = np.lexsort([M[:, i] for i in reversed(range(0, M.shape[1]))])
    S = M[I, :]

    i = 0

    while i < S.shape[0] - 1:
        l = []
        while np.all(S[i, :] == S[i + 1, :]):
            l.append(I[i])
            i += 1
        if len(l) > 0:
            l.append(I[i])
            res.append(l)
        i += 1

    return res


def cdist(A, B, **kwargs):
    if A.dtype != np.object:
        return scipy.spatial.distance.cdist(A, B, **kwargs)
    else:
        D = np.full((A.shape[0], B.shape[1]), np.inf, dtype=np.float)
        for i in range(A.shape[0]):
            for j in range(i + 1, B.shape[1]):
                d = M[i].distance_to(M[j])
                D[i, j], D[j, i] = d, d
        return D


if __name__ == '__main__':
    M = np.random.random((100, 3))

    M[3, :] = M[55, :]
    M[10, :] = M[55, :]

    print(get_duplicates(M))



import numpy as np

from pymoo.model.crossover import Crossover
from pymoo.rand import random


class SimulatedBinaryCrossover(Crossover):
    def __init__(self, prob_cross, eta_cross):
        super().__init__(2, 2)
        self.prob_cross = float(prob_cross)
        self.eta_cross = float(eta_cross)

    def _do(self, p, parents, children, D, **kwargs):
        n_matings = parents.shape[1]

        # crossover mask that will be used in the end
        do_crossover = np.full(parents[0].shape, True)

        # simulating probability of doing a crossover with the parents at all
        do_crossover[random.random(n_matings) > self.prob_cross, :] = False
        # per variable the probability is then 50%
        do_crossover[random.random((n_matings, p.n_var)) <= 0.5] = False
        # also if values are too close no mating is done
        do_crossover[np.abs(parents[0] - parents[1]) <= 1.0e-14] = False

        # assign y1 the smaller and y2 the larger value
        y1 = np.min(parents, axis=0)
        y2 = np.max(parents, axis=0)

        # random values for each individual
        rand = random.random((n_matings, p.n_var))

        def calc_betaq(beta):
            alpha = 2.0 - np.power(beta, -(self.eta_cross + 1.0))

            mask, mask_not = (rand <= (1.0 / alpha)), (rand > (1.0 / alpha))

            betaq = np.zeros(mask.shape)
            betaq[mask] = np.power((rand * alpha), (1.0 / (self.eta_cross + 1.0)))[mask]
            betaq[mask_not] = np.power((1.0 / (2.0 - rand * alpha)), (1.0 / (self.eta_cross + 1.0)))[mask_not]

            return betaq

        # difference between all variables
        delta = (y2 - y1)

        # now just be sure not dividing by zero (these cases will be filtered later anyway)
        delta[delta < 1.0e-14] = 1.0e-14

        beta = 1.0 + (2.0 * (y1 - p.xl) / delta)
        betaq = calc_betaq(beta)
        c1 = 0.5 * ((y1 + y2) - betaq * delta)

        beta = 1.0 + (2.0 * (p.xu - y2) / delta)
        betaq = calc_betaq(beta)
        c2 = 0.5 * ((y1 + y2) + betaq * delta)

        # do randomly a swap of variables
        b = random.random((n_matings, p.n_var)) <= 0.5
        val = c1[b]
        c1[b] = c2[b]
        c2[b] = val

        # take the parents as template
        c = parents

        # copy the positions where the crossover was done
        c[0, do_crossover] = c1[do_crossover]
        c[1, do_crossover] = c2[do_crossover]

        # copy to the structure which is returned
        children[:n_matings, :] = c[0]
        children[n_matings:, :] = c[1]

        # just be sure we are not out of bounds
        children[children < p.xl] = np.repeat(p.xl[None, :], children.shape[0], axis=0)[children < p.xl]
        children[children > p.xu] = np.repeat(p.xu[None, :], children.shape[0], axis=0)[children > p.xu]
